from datetime import timedelta
from glob import glob
from math import isnan
import os

import numpy as np
from osgeo import ogr, osr, gdal
from simpletail import ropen

from pthelma.cliapp import CliApp, WrongValueError
from pthelma.timeseries import Timeseries, datetime_from_iso, datestr_diff, \
    add_months_to_datetime


def idw(point, data_layer, alpha=1):
    data_layer.ResetReading()
    features = [f for f in data_layer if not isnan(f.GetField('value'))]
    distances = np.array([point.Distance(f.GetGeometryRef())
                          for f in features])
    values = np.array([f.GetField('value') for f in features])
    matches_station_exactly = abs(distances) < 1e-3
    if matches_station_exactly.any():
        invdistances = np.where(matches_station_exactly, 1, 0)
    else:
        invdistances = distances ** (-alpha)
    weights = invdistances / invdistances.sum()
    return (weights * values).sum()


def integrate(dataset, data_layer, target_band, funct, kwargs={}):
    mask = (dataset.GetRasterBand(1).ReadAsArray() != 0)

    # Create an array with the x co-ordinate of each grid point, and
    # one with the y co-ordinate of each grid point
    height, width = mask.shape
    x_left, x_step, d1, y_top, d2, y_step = dataset.GetGeoTransform()
    xcoords = np.arange(x_left + x_step / 2.0, x_left + x_step * width, x_step)
    ycoords = np.arange(y_top + y_step / 2.0, y_top + y_step * height, y_step)
    xarray, yarray = np.meshgrid(xcoords, ycoords)

    # Create a ufunc that makes the interpolation given the above arrays
    def interpolate_one_point(x, y, mask):
        if not mask:
            return np.nan
        point = ogr.Geometry(ogr.wkbPoint)
        point.AddPoint(x, y)
        return funct(point, data_layer, **kwargs)
    interpolate = np.vectorize(interpolate_one_point,
                               otypes=[np.float32, np.float32, np.bool])

    # Make the calculation
    target_band.WriteArray(interpolate(xarray, yarray, mask))


def create_ogr_layer_from_timeseries(filenames, epsg, data_source):
    # Prepare the co-ordinate transformation from WGS84 to epsg
    source_sr = osr.SpatialReference()
    source_sr.ImportFromEPSG(4326)
    target_sr = osr.SpatialReference()
    target_sr.ImportFromEPSG(epsg)
    transform = osr.CoordinateTransformation(source_sr, target_sr)

    layer = data_source.CreateLayer('stations', target_sr)
    layer.CreateField(ogr.FieldDefn('filename', ogr.OFTString))
    for filename in filenames:
        with open(filename) as f:
            ts = Timeseries()
            ts.read_meta(f)
        point = ogr.Geometry(ogr.wkbPoint)
        point.AddPoint(ts.location['abscissa'], ts.location['ordinate'])
        point.Transform(transform)
        f = ogr.Feature(layer.GetLayerDefn())
        f.SetGeometry(point)
        f.SetField('filename', filename)
        layer.CreateFeature(f)
    return layer


def h_integrate(mask, stations_layer, date, output_filename, date_fmt,
                funct, kwargs):
    # Return immediately if output file already exists
    if os.path.exists(output_filename):
        return

    # Read the time series values and add the 'value' attribute to
    # stations_layer
    stations_layer.CreateField(ogr.FieldDefn('value', ogr.OFTReal))
    found_value = False
    stations_layer.ResetReading()
    for station in stations_layer:
        filename = station.GetField('filename')
        t = Timeseries()
        with open(filename) as f:
            t.read_file(f)
        value = t.get(date, float('NaN'))
        station.SetField('value', value)
        found_value = found_value or (not isnan(value))
        stations_layer.SetFeature(station)
    if not found_value:
        return

    # Create destination data source
    output = gdal.GetDriverByName('GTiff').Create(
        output_filename, mask.RasterXSize, mask.RasterYSize, 1,
        gdal.GDT_Float32)
    if not output:
        raise IOError('An error occured when trying to open {}'.format(
            output_filename))
    output.SetMetadataItem('TIMESTAMP', date.strftime(date_fmt))

    try:
        # Set geotransform and projection in the output data source
        output.SetGeoTransform(mask.GetGeoTransform())
        output.SetProjection(mask.GetProjection())

        # Do the integration
        integrate(mask, stations_layer, output.GetRasterBand(1), funct, kwargs)
    finally:
        # Close the dataset
        output = None


class BitiaApp(CliApp):
    name = 'bitia'
    description = 'Perform spatial integration'
                        # Section          Option            Default
    config_file_options = {'General': {'mask':             None,
                                       'epsg':             None,
                                       'output_dir':       None,
                                       'filename_prefix':  None,
                                       'files_to_produce': None,
                                       'method':           None,
                                       'alpha':            '1',
                                       'files':            None,
                                       },
                           }

    def read_configuration(self):
        super(BitiaApp, self).read_configuration()
        self.files = self.config['General']['files'].split('\n')

    def check_configuration(self):
        super(BitiaApp, self).check_configuration()
        self.check_configuration_method()
        self.check_configuration_epsg()

    def check_configuration_method(self):
        # Check method
        if self.config['General']['method'] != 'idw':
            raise WrongValueError('Option "method" can currently only be idw')
        # Check alpha
        try:
            float(self.config['General']['alpha'])
        except ValueError:
            raise WrongValueError('Option "alpha" must be a number')

    def check_configuration_epsg(self):
        try:
            epsg = int(self.config['General']['epsg'])
        except ValueError:
            raise WrongValueError('Option "epsg" must be an integer')
        srs = osr.SpatialReference()
        result = srs.ImportFromEPSG(epsg)
        if result:
            raise WrongValueError(
                'An error occurred when trying to use epsg={}'.format(epsg))

    @property
    def last_date(self):
        """
        Return the last date for which any time series has data, or None if
        all time series are empty.
        """
        result = None
        for filename in self.files:
            last_date = None
            if not os.path.exists(filename):
                continue
            with ropen(filename) as fp:
                for line in fp:
                    if not line.strip():
                        break  # Time series has no data
                    datestring = line.split(',')[0]
                    try:
                        last_date = datetime_from_iso(datestring)
                    except ValueError as e:
                        raise ValueError(e.message + ' (file {}, last line)'
                                         .format(filename))
                    break  # We only want to do this for the last line
            if last_date and ((not result) or (last_date > result)):
                result = last_date
        return result

    @property
    def time_step(self):
        """
        Return time step of all time series. If time step is not the same
        for all time series, raises exception.
        """
        time_step = None
        for filename in self.files:
            with open(filename) as f:
                t = Timeseries()
                t.read_meta(f)
            item_time_step = (t.time_step.length_minutes,
                              t.time_step.length_months)
            if time_step and (item_time_step != time_step):
                raise WrongValueError(
                    'Not all time series have the same step')
            time_step = item_time_step
        return time_step

    @property
    def date_fmt(self):
        """
        Determine date_fmt based on time series time step.
        """
        minutes, months = self.time_step
        if minutes and (minutes < 1440):
            return '%Y-%m-%d %H:%M'
        if minutes and (minutes >= 1440):
            return '%Y-%m-%d'
        if months and (months < 12):
            return '%Y-%m'
        if months and (months >= 12):
            return '%Y'
        raise ValueError("Can't use time step " + str(self.time_step))

    def rename_existing_files(self, last_date):
        """
        Assume filename_prefix-XXXX files exist, where XXXX goes from 0 to N-1,
        where N is the 'files_to_produce' configuration option. This method
        renames (by increasing XXXX) so that if a new file with XXXX=000 is
        created for the specified last_date, the existing ones will be properly
        numbered (usually XXXX is increased by one, but it can be more if the
        system failed to run for some time steps).  It also removes files with
        resulting XXXX >= N.
        """
        output_dir = self.config['General']['output_dir']
        filename_prefix = self.config['General']['filename_prefix']
        pattern = os.path.join(output_dir, '{}-*.tif'.format(filename_prefix))
        files = glob(pattern)
        files.sort(reverse=True)
        for filename in files:
            self.do_rename_file(filename, last_date)

    def do_rename_file(self, filename, last_date):
        # Get the file's timestamp
        fp = gdal.Open(filename)
        timestamp = fp.GetMetadata()['TIMESTAMP']
        fp = None

        # What should the file's number be?
        number = self.get_file_number(last_date, timestamp)

        # Is the file too old? Delete it.
        if number >= int(self.config['General']['files_to_produce']):
            os.unlink(filename)
            return

        # Get the file's current number and rename the file
        cur_number = int(filename.rpartition('-')[2].split('.')[0])
        if number != cur_number:
            new_filename = os.path.join(
                self.config['General']['output_dir'],
                '{}-{:04d}.tif'.format(
                    self.config['General']['filename_prefix'], number))
            os.rename(filename, new_filename)

    def get_file_number(self, last_date, timestamp, filename=''):
        """
        Return the number a file should have for the given last_date and
        timestamp. If the two days are equal, the result is 0; otherwise
        it is the number of time steps between them, or 10000 if the
        difference is too large. The filename is used for error messages.
        """
        diff_months, diff_minutes = datestr_diff(last_date, timestamp)
        step_minutes, step_months = self.time_step
        if step_months:  # Monthly or annual
            number = int(diff_months / step_months)
            if diff_minutes or (number * step_months != diff_months):
                raise ValueError("Something's wrong in the timestamp in {}"
                                 .format(filename))
        else:
            number = int(diff_minutes / step_minutes)
            if number * step_minutes != diff_minutes:
                raise ValueError("Something's wrong in the timestamp in {}"
                                 .format(filename))
            if diff_months:
                number = 10000
        return number

    def execute(self):
        # Create stations layer
        stations = ogr.GetDriverByName('memory').CreateDataSource('stations')
        epsg = int(self.config['General']['epsg'])
        stations_layer = create_ogr_layer_from_timeseries(self.files, epsg,
                                                          stations)

        # Get mask
        mask = gdal.Open(self.config['General']['mask'])
        if mask is None:
            raise IOError('An error occured when trying to open {}'.format(
                self.config['General']['mask']))

        # Setup integration method
        if self.config['General']['method'] == 'idw':
            funct = idw
            kwargs = {'alpha': float(self.config['General']['alpha']), }
        else:
            assert False

        # Rename existing files and remove old files
        last_date_str = self.last_date.strftime(self.date_fmt)
        self.rename_existing_files(last_date_str)

        # Make calculation for each missing file
        output_dir = self.config['General']['output_dir']
        filename_prefix = self.config['General']['filename_prefix']
        date_to_calculate = self.last_date
        while True:
            dtc_str = date_to_calculate.strftime(self.date_fmt)
            number = self.get_file_number(last_date_str, dtc_str)
            if number >= int(self.config['General']['files_to_produce']):
                break
            filename = os.path.join(
                output_dir, '{}-{:04d}.tif'.format(filename_prefix, number))
            h_integrate(mask, stations_layer, date_to_calculate, filename,
                        self.date_fmt, funct, kwargs)
            step_minutes, step_months = self.time_step
            if step_months:
                date_to_calculate = add_months_to_datetime(date_to_calculate,
                                                           -step_months)
            else:
                date_to_calculate = date_to_calculate - timedelta(
                    minutes=step_minutes)



"""
In order to run the meteologger tests, you must specify the
PTHELMA_TEST_ENHYDRIS_API environment variable as explained in
test_enhydris_api.py.
"""

from datetime import datetime
import json
from math import isnan
import os
import shutil
import sys
import tempfile
import textwrap
from unittest import TestCase, skipIf, skipUnless

from six import StringIO
from six.moves.configparser import NoOptionError

import pytz

from pthelma import enhydris_api
from pthelma.cliapp import WrongValueError, InvalidOptionError
from pthelma.meteologger import Datafile_deltacom, Datafile_simple, \
    Datafile_wdat5, Datafile_odbc, ConfigurationError, LoggertodbApp
from pthelma.timeseries import Timeseries


def full_testdata_filename(filename):
    current_directory = os.path.dirname(os.path.abspath(__file__))
    full_pathname = os.path.join(current_directory, 'data', filename)
    return full_pathname


def create_timeseries(cookies, adict):
    # Create a timeseries on the server and return its id
    j = {'gentity': adict['station_id'],
         'variable': adict['variable_id'],
         'unit_of_measurement': adict['unit_of_measurement_id'],
         'time_zone': adict['time_zone_id'], }
    return enhydris_api.post_model(adict['base_url'], cookies, 'Timeseries', j)


def get_server_from_env(adict):
    conn = os.getenv("PTHELMA_TEST_ENHYDRIS_API")
    v = json.loads(conn)
    for item in 'base_url user password station_id variable_id ' \
                'unit_of_measurement_id time_zone_id'.split():
        adict[item] = v[item]


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            "set PTHELMA_TEST_ENHYDRIS_API")
class _Test_logger(TestCase):
    class_being_tested = None
    datafilename = ''
    datafiledict = {}
    ref_ts1 = Timeseries(0)
    ref_ts1.read(open(full_testdata_filename('timeseries1.txt')))
    ref_ts2 = Timeseries(0)
    ref_ts2.read(open(full_testdata_filename('timeseries2.txt')))

    def setUp(self):
        get_server_from_env(self.__dict__)

        if not self.class_being_tested or not self.base_url:
            return

        # First 60 nonempty lines of test go to self.file1
        (fd, self.file1) = tempfile.mkstemp(text=True)
        fp = os.fdopen(fd, 'w')
        i = 0
        with open(full_testdata_filename(self.datafilename)) as f:
            for line in f:
                if i >= 60:
                    break
                if not line.strip():
                    continue
                i += 1
                fp.write(line)
        fp.close()

        # Login
        self.cookies = enhydris_api.login(self.base_url,
                                          self.user,
                                          self.password)

        # Create two timeseries
        self.timeseries_id1 = create_timeseries(self.cookies, self.__dict__)
        self.timeseries_id2 = create_timeseries(self.cookies, self.__dict__)
        self.datafile_fields = "0,{0},{1}".format(self.timeseries_id1,
                                                  self.timeseries_id2)
        self.ts1 = Timeseries(self.timeseries_id1)
        self.ts2 = Timeseries(self.timeseries_id2)

    def tearDown(self):
        if not self.class_being_tested or not self.base_url:
            return
        enhydris_api.delete_model(self.base_url, self.cookies,
                                  'Timeseries', self.timeseries_id2)
        enhydris_api.delete_model(self.base_url, self.cookies,
                                  'Timeseries', self.timeseries_id1)

    def check_config_test(self):
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.cookies, {})
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.cookies,
                          {'filename': 'hello', 'datafile_fields': '0',
                           'datafile_format': 'irrelevant',
                           'nonexistent_config_option': True})
        # Call it correctly and expect it doesn't raise anything
        self.class_being_tested(self.base_url, self.cookies,
                                {'filename': 'hello', 'datafile_fields': '0',
                                 'datafile_format': 'irrelevant'})

    def upload_test(self):
        d = {'filename': self.file1,
             'datafile_fields': self.datafile_fields,
             'datafile_format': 'irrelevant'}
        d.update(self.datafiledict)
        df = self.class_being_tested(self.base_url, self.cookies, d)
        df.update_database()
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts1)
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts2)
        self.assertEqual(len(self.ts1), 60)
        self.assertEqual(len(self.ts2), 60)
        (items1, items2, ritems1, ritems2) = [
            x.items() for x in (self.ts1, self.ts2,
                                self.ref_ts1, self.ref_ts2)]
        for i in range(0, 60):
            self.assertEqual(items1[i][0], ritems1[i][0])
            self.assertAlmostEqual(items1[i][1], ritems1[i][1], 4)
            self.assertEqual(items1[i][1].flags, ritems1[i][1].flags)
            self.assertEqual(items2[i][0], ritems2[i][0])
            self.assertAlmostEqual(items2[i][1], ritems2[i][1], 4)
            self.assertEqual(items2[i][1].flags, ritems2[i][1].flags)
        d['filename'] = full_testdata_filename(self.datafilename)
        df = self.class_being_tested(self.base_url, self.cookies, d)
        df.update_database()
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts1)
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts2)
        self.assertEqual(len(self.ts1), 100)
        self.assertEqual(len(self.ts2), 100)
        (items1, items2, ritems1, ritems2) = [
            x.items() for x in (self.ts1, self.ts2,
                                self.ref_ts1, self.ref_ts2)]
        for i in range(0, 100):
            self.assertEqual(items1[i][0], ritems1[i][0])
            self.assertAlmostEqual(items1[i][1], ritems1[i][1], 4)
            self.assertEqual(items1[i][1].flags, ritems1[i][1].flags)
            self.assertEqual(items2[i][0], ritems2[i][0])
            self.assertAlmostEqual(items2[i][1], ritems2[i][1], 4)
            self.assertEqual(items2[i][1].flags, ritems2[i][1].flags)

    def runTest(self):
        if not self.class_being_tested or not self.base_url:
            return
        self.check_config_test()
        self.upload_test()


class TestDeltacom(_Test_logger):
    datafilename = 'deltacom_data.txt'
    class_being_tested = Datafile_deltacom


class TestSimple1(_Test_logger):
    datafilename = 'simple_data1.txt'
    class_being_tested = Datafile_simple
    datafiledict = {'date_format': '%y/%m/%d %H:%M:%S'}


class TestSimple2(_Test_logger):
    datafilename = 'simple_data2.txt'
    class_being_tested = Datafile_simple
    datafiledict = {'date_format': '%d/%m/%Y %H:%M:%S'}


class TestSimple3(_Test_logger):
    datafilename = 'simple_data3.txt'
    class_being_tested = Datafile_simple
    datafiledict = {'date_format': '%d/%m/%Y %H:%M',
                    'nfields_to_ignore': 1,
                    'delimiter': ','}


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            "set PTHELMA_TEST_ENHYDRIS_API")
class TestDst(TestCase):
    datafiledict = {'date_format': '%Y-%m-%dT%H:%M',
                    'timezone': 'Europe/Athens'}

    def setUp(self):
        get_server_from_env(self.__dict__)
        self.ref_ts = Timeseries(0)
        if not self.base_url:
            return
        self.cookies = enhydris_api.login(self.base_url, self.user,
                                          self.password)
        self.timeseries_id = create_timeseries(self.cookies, self.__dict__)
        self.ts = Timeseries(self.timeseries_id)

    def tearDown(self):
        if not self.base_url:
            return
        enhydris_api.delete_model(self.base_url, self.cookies,
                                  'Timeseries', self.timeseries_id)

    def run_test(self):
        if not self.base_url:
            return
        d = {'filename': full_testdata_filename(self.filename),
             'datafile_fields': str(self.timeseries_id),
             'datafile_format': 'irrelevant'}
        d.update(self.datafiledict)
        df = Datafile_simple(self.base_url, self.cookies, d)
        df.update_database()
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts)
        self.assertEqual(len(self.ts), len(self.ref_ts))
        (items, ritems) = [x.items() for x in (self.ts, self.ref_ts)]
        for item, ritem in zip(items, ritems):
            self.assertEqual(item[0], ritem[0])
            self.assertAlmostEqual(item[1], ritem[1], 4)
            self.assertEqual(item[1].flags, ritem[1].flags)

    @skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'), "see above")
    def test_to_dst(self):
        self.filename = 'data_at_change_to_dst.txt'
        with open(full_testdata_filename('timeseries_at_change_to_dst.txt')
                  ) as f:
            self.ref_ts.read(f)
        self.run_test()

    @skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'), "see above")
    def test_from_dst(self):
        self.filename = 'data_at_change_from_dst.txt'
        with open(full_testdata_filename('timeseries_at_change_from_dst.txt')
                  ) as f:
            self.ref_ts.read(f)
        self.run_test()

    def test_fix_dst(self):
        d = {'filename': 'irrelevant',
             'datafile_fields': '0',
             'datafile_format': 'irrelevant'}
        d.update(self.datafiledict)
        df = Datafile_simple('http://irrelevant/', {}, d)
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 2, 59)),
                         datetime(2012, 10, 28, 1, 59))
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 3, 00)),
                         datetime(2012, 10, 28, 3, 00))
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 4, 00)),
                         datetime(2012, 10, 28, 4, 00))

        # Now we pretend that the switch from dst hasn't occurred yet.
        # This is the only case when loggertodb should assume that
        # ambiguous times refer to before the switch.
        athens = pytz.timezone('Europe/Athens')
        now = athens.localize(datetime(2012, 10, 28, 3, 59), is_dst=True)
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 2, 59), now=now),
                         datetime(2012, 10, 28, 1, 59))
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 3, 00), now=now),
                         datetime(2012, 10, 28, 2, 00))
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 4, 00), now=now),
                         datetime(2012, 10, 28, 4, 00))

        # Once more; the switch from DST has just occurred; now it
        # should be assumed that ambiguous times refer to after the
        # switch.
        now = athens.localize(datetime(2012, 10, 28, 3, 0), is_dst=False)
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 2, 59), now=now),
                         datetime(2012, 10, 28, 1, 59))
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 3, 00), now=now),
                         datetime(2012, 10, 28, 3, 00))
        self.assertEqual(df._fix_dst(datetime(2012, 10, 28, 4, 00), now=now),
                         datetime(2012, 10, 28, 4, 00))


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            "set PTHELMA_TEST_ENHYDRIS_API")
class TestWdat5(_Test_logger):
    class_being_tested = Datafile_wdat5
    longMessage = True
    # The dicts below hold the parameter name as it is in the WeatherLink
    # README file, and the corresponding heading in the WeatherLink export file
    # (lowercase, after changing spaces to hyphens and removing dots), empty
    # string for those that aren't going to be tested.
    parameters = [
        {'name': 'outsideTemp', 'expname': 'temp-out'},
        {'name': 'hiOutsideTemp', 'expname': 'hi-temp'},
        {'name': 'lowOutsideTemp', 'expname': 'low-temp'},
        {'name': 'insideTemp', 'expname': 'in-temp'},
        {'name': 'barometer', 'expname': 'bar'},
        {'name': 'outsideHum', 'expname': 'out-hum'},
        {'name': 'insideHum', 'expname': 'in-hum'},
        {'name': 'rain', 'expname': 'rain'},
        {'name': 'hiRainRate', 'expname': 'rain-rate'},
        {'name': 'windSpeed', 'expname': 'wind-speed'},
        {'name': 'hiWindSpeed', 'expname': 'hi-speed'},
        {'name': 'windDirection', 'expname': 'wind-dir'},
        {'name': 'hiWindDirection', 'expname': 'hi-dir'},
        {'name': 'numWindSamples', 'expname': 'wind-samp'},
        {'name': 'solarRad', 'expname': 'solar-rad'},
        {'name': 'hiSolarRad', 'expname': 'hi-solar-rad'},
        {'name': 'UV', 'expname': 'uv-index'},
        {'name': 'hiUV', 'expname': 'hi-uv'},
        {'name': 'leafTemp1', 'expname': ''},
        {'name': 'leafTemp2', 'expname': ''},
        {'name': 'leafTemp3', 'expname': ''},
        {'name': 'leafTemp4', 'expname': ''},
        {'name': 'extraRad', 'expname': ''},
        {'name': 'newSensors1', 'expname': ''},
        {'name': 'newSensors2', 'expname': ''},
        {'name': 'newSensors3', 'expname': ''},
        {'name': 'newSensors4', 'expname': ''},
        {'name': 'newSensors5', 'expname': ''},
        {'name': 'newSensors6', 'expname': ''},
        {'name': 'forecast', 'expname': ''},
        {'name': 'ET', 'expname': 'et'},
        {'name': 'soilTemp1', 'expname': ''},
        {'name': 'soilTemp2', 'expname': ''},
        {'name': 'soilTemp3', 'expname': ''},
        {'name': 'soilTemp4', 'expname': ''},
        {'name': 'soilTemp5', 'expname': ''},
        {'name': 'soilTemp6', 'expname': ''},
        {'name': 'soilMoisture1', 'expname': ''},
        {'name': 'soilMoisture2', 'expname': ''},
        {'name': 'soilMoisture3', 'expname': ''},
        {'name': 'soilMoisture4', 'expname': ''},
        {'name': 'soilMoisture5', 'expname': ''},
        {'name': 'soilMoisture6', 'expname': ''},
        {'name': 'leafWetness1', 'expname': ''},
        {'name': 'leafWetness2', 'expname': ''},
        {'name': 'leafWetness3', 'expname': ''},
        {'name': 'leafWetness4', 'expname': ''},
        {'name': 'extraTemp1', 'expname': ''},
        {'name': 'extraTemp2', 'expname': ''},
        {'name': 'extraTemp3', 'expname': ''},
        {'name': 'extraTemp4', 'expname': ''},
        {'name': 'extraTemp5', 'expname': ''},
        {'name': 'extraTemp6', 'expname': ''},
        {'name': 'extraTemp7', 'expname': ''},
        {'name': 'extraHum1', 'expname': ''},
        {'name': 'extraHum2', 'expname': ''},
        {'name': 'extraHum3', 'expname': ''},
        {'name': 'extraHum4', 'expname': ''},
        {'name': 'extraHum5', 'expname': ''},
        {'name': 'extraHum6', 'expname': ''},
        {'name': 'extraHum7', 'expname': ''},
    ]

    def setUp(self):
        get_server_from_env(self.__dict__)

        # Connect to server
        self.cookies = enhydris_api.login(self.base_url, self.user,
                                          self.password)

        # Create as many timeseries as there are not-to-ignore parameters
        for parm in self.parameters:
            parm['ts_id'] = create_timeseries(self.cookies, self.__dict__
                                              ) if parm['expname'] else 0

    def tearDown(self):
        for parm in self.parameters:
            if parm['ts_id']:
                enhydris_api.delete_model(self.base_url, self.cookies,
                                          'Timeseries', parm['ts_id'])

    def check_config_test(self):
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.cookies, {})
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.cookies,
                          {'filename': 'hello', 'outsidetemp': '0',
                           'datafile_format': 'irrelevant',
                           'nonexistent_config_option': True})
        # Call it correctly and expect it doesn't raise anything
        self.class_being_tested(self.base_url, self.cookies,
                                {'filename': 'hello', 'outsidetemp': '0',
                                 'datafile_format': 'irrelevant'})

    def upload_test(self):
        d = {'filename': full_testdata_filename(os.path.join('wdat5', '1')),
             'datafile_format': 'irrelevant'}
        for parm in self.parameters:
            if parm['ts_id']:
                d[parm['name'].lower()] = parm['ts_id']
        df = Datafile_wdat5(self.base_url, self.cookies, d)
        df.update_database()
        self.check(d['filename'])
        d['filename'] = full_testdata_filename(os.path.join('wdat5', '2'))
        df = Datafile_wdat5(self.base_url, self.cookies, d)
        df.update_database()
        self.check(d['filename'])

    def check(self, datadir):
        for parm in self.parameters:
            if not parm['ts_id']:
                continue
            actual_ts = Timeseries(parm['ts_id'])
            enhydris_api.read_tsdata(self.base_url, self.cookies, actual_ts)
            reference_ts = Timeseries()
            with open(os.path.join(
                    datadir, 'generated', parm['expname'] + '.txt')) as f:
                reference_ts.read(f)
                precision = self.guess_precision(f)
            self.assertTimeseriesEqual(actual_ts, reference_ts, precision,
                                       parm['expname'] + '.txt')

    def guess_precision(self, f):
        result = 0
        f.seek(0)
        # Take a sample of 200 rows
        for i, line in enumerate(f):
            if i >= 200:
                break
            value = line.split(',')[1].strip()
            integral, sep, decimal = value.partition('.')
            result = len(decimal) if len(decimal) > result else result
        return result

    def assertTimeseriesEqual(self, ts1, ts2, precision, filename):
        self.assertEqual(len(ts1), len(ts2))
        items1 = ts1.items()
        items2 = ts2.items()
        for i, (item1, item2) in enumerate(zip(items1, items2)):
            self.assertEqual(item1[0], item2[0])
            if isnan(item1[1]):
                self.assertTrue(isnan(item2[1]),
                                msg='{1}, row {0}'.format(i + 1, filename))
            else:
                self.assertAlmostEqual(
                    item1[1], item2[1], places=precision,
                    msg='{1}, row {0}'.format(i + 1, filename))
            self.assertEqual(item1[1].flags, item2[1].flags)

    def assertAlmostEqual(self, a, b, places=7, msg=None):
        """We redefine assertAlmostEqual so that, e.g. with places=1, 0.55 is
        considered equal to both 0.5 and 0.6
        """
        tolerance = 10 ** (-places - 3)
        if abs(a - b) <= 0.5 * 10 ** (-places) + tolerance:
            return
        super(TestWdat5, self).assertAlmostEqual(a, b, places=places, msg=msg)


skip_test_odbc = True
skip_test_odbc_message = ''
if sys.platform != 'win32':
    skip_test_odbc_message = 'Windows only'
elif not os.getenv('PTHELMA_TEST_ENHYDRIS_API'):
    skip_test_odbc_message = 'set PTHELMA_TEST_ENHYDRIS_API'
else:
    try:
        import pyodbc
        if 'Microsoft Access Driver (*.mdb)' not in pyodbc.drivers():
            skip_test_odbc_message = \
                'Install ODBC "Microsoft Access Driver (*.mdb)"'
        else:
            skip_test_odbc = False
    except ImportError:
        skip_test_odbc_message = 'Install pyodbc'


@skipIf(skip_test_odbc, skip_test_odbc_message)
class TestOdbc(TestCase):
    class_being_tested = Datafile_odbc
    ref_ts1 = Timeseries(0)
    ref_ts1.read(StringIO(textwrap.dedent('''\
                                          2014-03-30 10:55,12.8,
                                          2014-03-30 10:56,12.8,
                                          2014-03-30 10:57,12.9,
                                          2014-03-30 23:12,12.9,
                                          2014-03-30 23:42,12.9,
                                          2014-03-31 00:12,11.9,
                                          2014-03-31 00:27,11.7,
                                          2014-03-31 00:42,11.7,
                                          2014-03-31 01:27,10.9,
                                          2014-03-31 15:11,21,
                                          2014-03-31 19:26,16.4,
                                          2014-03-31 20:11,14.6,
                                          2014-03-31 20:26,14.3,
                                          2014-03-31 20:41,13.9,
                                          2014-04-01 14:41,20.8,
                                          2014-04-01 15:41,21,
                                          2014-04-01 15:56,21.3,
                                          ''')))
    ref_ts2 = Timeseries(0)
    ref_ts2.read(StringIO(textwrap.dedent('''\
                                          2014-03-30 10:55,99.9,
                                          2014-03-30 10:56,99.8,
                                          2014-03-30 10:57,99.9,
                                          2014-03-30 23:12,99.9,
                                          2014-03-30 23:42,99.9,
                                          2014-03-31 00:12,99.8,
                                          2014-03-31 00:27,99.8,
                                          2014-03-31 00:42,99.9,
                                          2014-03-31 01:27,99.8,
                                          2014-03-31 15:11,52.3,
                                          2014-03-31 19:26,80.8,
                                          2014-03-31 20:11,87.7,
                                          2014-03-31 20:26,91,
                                          2014-03-31 20:41,93.6,
                                          2014-04-01 14:41,78,
                                          2014-04-01 15:41,73.8,
                                          2014-04-01 15:56,66.1,
                                          ''')))
    file1 = 'DRIVER=Microsoft Access Driver (*.mdb);DBQ={}'.format(
        full_testdata_filename('msaccess1.mdb'))
    file2 = 'DRIVER=Microsoft Access Driver (*.mdb);DBQ={}'.format(
        full_testdata_filename('msaccess2.mdb'))
    datafiledict = {'table': 'Clima',
                    'date_sql': "Date + ' ' + Time",
                    'data_columns': 'T out,H out',
                    'date_format': '%d/%m/%Y %H:%M:%S',
                    'decimal_separator': ',',
                    }

    def setUp(self):
        get_server_from_env(self.__dict__)

        if not self.class_being_tested or not self.base_url:
            return

        # Login
        self.cookies = enhydris_api.login(self.base_url,
                                          self.user,
                                          self.password)

        # Create two timeseries
        self.timeseries_id1 = create_timeseries(self.cookies, self.__dict__)
        self.timeseries_id2 = create_timeseries(self.cookies, self.__dict__)
        self.datafile_fields = "{0},{1}".format(self.timeseries_id1,
                                                self.timeseries_id2)
        self.ts1 = Timeseries(self.timeseries_id1)
        self.ts2 = Timeseries(self.timeseries_id2)

    def tearDown(self):
        if not self.class_being_tested or not self.base_url:
            return
        enhydris_api.delete_model(self.base_url, self.cookies,
                                  'Timeseries', self.timeseries_id2)
        enhydris_api.delete_model(self.base_url, self.cookies,
                                  'Timeseries', self.timeseries_id1)

    def check_config_test(self):
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.cookies, {})
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.cookies,
                          {'filename': 'hello', 'datafile_fields': '0',
                           'datafile_format': 'msaccess',
                           'table': 'atable',
                           'date_sql': 'somecolumn',
                           'data_columns': 'acolumn1,acolumn2',
                           'nonexistent_config_option': True})
        # Call it correctly and expect it doesn't raise anything
        self.class_being_tested(self.base_url, self.cookies,
                                {'filename': 'hello', 'datafile_fields': '0',
                                 'datafile_format': 'msaccess',
                                 'table': 'atable',
                                 'date_sql': 'somecolumn',
                                 'data_columns': 'acolumn1,acolumn2',
                                 })

    def upload_test(self):
        d = {'filename': self.file1,
             'datafile_fields': self.datafile_fields,
             'datafile_format': 'msaccess',
             'table': 'Clima'}
        d.update(self.datafiledict)
        df = self.class_being_tested(self.base_url, self.cookies, d)
        df.update_database()
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts1)
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts2)
        self.assertEqual(len(self.ts1), 10)
        self.assertEqual(len(self.ts2), 10)
        (items1, items2, ritems1, ritems2) = [
            x.items() for x in (self.ts1, self.ts2,
                                self.ref_ts1, self.ref_ts2)]
        for i in range(0, 10):
            self.assertEqual(items1[i][0], ritems1[i][0])
            self.assertAlmostEqual(items1[i][1], ritems1[i][1], 4)
            self.assertEqual(items1[i][1].flags, ritems1[i][1].flags)
            self.assertEqual(items2[i][0], ritems2[i][0])
            self.assertAlmostEqual(items2[i][1], ritems2[i][1], 4)
            self.assertEqual(items2[i][1].flags, ritems2[i][1].flags)
        d['filename'] = self.file2
        df = self.class_being_tested(self.base_url, self.cookies, d)
        df.update_database()
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts1)
        enhydris_api.read_tsdata(self.base_url, self.cookies, self.ts2)
        self.assertEqual(len(self.ts1), 17)
        self.assertEqual(len(self.ts2), 17)
        (items1, items2, ritems1, ritems2) = [
            x.items() for x in (self.ts1, self.ts2,
                                self.ref_ts1, self.ref_ts2)]
        for i in range(0, 17):
            self.assertEqual(items1[i][0], ritems1[i][0])
            self.assertAlmostEqual(items1[i][1], ritems1[i][1], 4)
            self.assertEqual(items1[i][1].flags, ritems1[i][1].flags)
            self.assertEqual(items2[i][0], ritems2[i][0])
            self.assertAlmostEqual(items2[i][1], ritems2[i][1], 4)
            self.assertEqual(items2[i][1].flags, ritems2[i][1].flags)

    def runTest(self):
        if not self.class_being_tested or not self.base_url:
            return
        self.check_config_test()
        self.upload_test()


class LoggertodbAppTestCase(TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config_file = os.path.join(self.tempdir, 'loggertodb.conf')
        self.saved_argv = sys.argv
        sys.argv = ['loggertodb', '--traceback', self.config_file]

    def tearDown(self):
        sys.argv = self.saved_argv
        shutil.rmtree(self.tempdir)

    def test_nonexistent_config_file(self):
        app = LoggertodbApp()
        sys.argv = ['loggertodb',
                    os.path.join(self.tempdir, 'nonexistent.conf')]
        stderr = StringIO()
        orig_stderr = sys.stderr
        sys.stderr = stderr
        try:
            app.run()
            self.assertTrue(False)
        except SystemExit as e:
            self.assertTrue('nonexistent.conf' in stderr.getvalue())
        finally:
            sys.stderr = orig_stderr

    def test_wrong_configuration1(self):
        app = LoggertodbApp()
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                user = a_user
                password = a_password
                '''))
        self.assertRaises(NoOptionError, app.run, dry=True)

    def test_wrong_configuration2(self):
        app = LoggertodbApp()
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_url = a_base_url
                user = a_user
                password = a_password
                nonexistent_option = an_option
                '''))
        self.assertRaises(InvalidOptionError, app.run, dry=True)

    def test_wrong_configuration3(self):
        app = LoggertodbApp()
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_url = a_base_url
                user = a_user
                password = a_password
                loglevel = NONEXISTENT_LOG_LEVEL
                '''))
        self.assertRaises(WrongValueError, app.run, dry=True)

    def test_correct_configuration1(self):
        app = LoggertodbApp()
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_url = a_base_url
                user = a_user
                password = a_password
                '''))
        app.run(dry=True)

    def test_correct_configuration2(self):
        app = LoggertodbApp()
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_url = a_base_url
                user = a_user
                password = a_password
                loglevel = ERROR
                '''))
        app.run(dry=True)



import os
import shutil
from six import StringIO
from six.moves import configparser
import sys
import tempfile
import textwrap
from unittest import TestCase

from pthelma.cliapp import WrongValueError
from pthelma.fordonia import FordoniaApp
from pthelma.timeseries import Timeseries


class FordoniaAppTestCase(TestCase):

    def __init__(self, *args, **kwargs):
        super(FordoniaAppTestCase, self).__init__(*args, **kwargs)

        # Python 2.7 compatibility
        try:
            self.assertRaisesRegex
        except AttributeError:
            self.assertRaisesRegex = self.assertRaisesRegexp

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config_file = os.path.join(self.tempdir, 'fordonia.conf')
        self.saved_argv = sys.argv
        sys.argv = ['fordonia', '--traceback', self.config_file]
        self.savedcwd = os.getcwd()

        # Create two time series
        self.filenames = [os.path.join(self.tempdir, x)
                          for x in ('ts1', 'ts2')]
        self.output_filenames = [x + 'a' for x in self.filenames]
        with open(self.filenames[0], 'w') as f:
            f.write("Location=23.78743 37.97385 4326\n"
                    "Time_step=10,0\n"
                    "Nominal_offset=0,0\n"
                    "Actual_offset=0,0\n"
                    "\n"
                    "2014-06-16 14:50,14.1,\n"
                    "2014-06-16 15:00,15.0,\n"
                    "2014-06-16 15:10,16.9,\n"
                    "2014-06-16 15:20,17.8,\n"
                    "2014-06-16 15:30,18.7,\n"
                    "2014-06-16 15:40,19.6,\n"
                    "2014-06-16 15:50,20.5,\n"
                    "2014-06-16 16:00,21.4,\n"
                    "2014-06-16 16:10,22.3,\n"
                    )
        with open(self.filenames[1], 'w') as f:
            f.write("Location=24.56789 38.76543 4326\n"
                    "Time_step=10,0\n"
                    "Nominal_offset=0,0\n"
                    "Actual_offset=0,0\n"
                    "\n"
                    "2014-06-17 14:50,94.1,\n"
                    "2014-06-17 15:00,85.0,\n"
                    "2014-06-17 15:10,76.9,\n"
                    "2014-06-17 15:20,67.8,\n"
                    "2014-06-17 15:30,58.7,\n"
                    "2014-06-17 15:40,49.6,\n"
                    "2014-06-17 15:50,30.5,\n"
                    "2014-06-17 16:00,21.4,\n"
                    "2014-06-17 16:10,12.3,\n"
                    )

        # Prepare a configuration file (some tests override it)
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_dir = {self.tempdir}
                target_step = 60, 0

                [timeseries1]
                source_file = ts1
                target_file = ts1a
                interval_type = sum

                [timeseries2]
                source_file = ts2
                target_file = ts2a
                interval_type = average
                ''').format(self=self))

    def tearDown(self):
        os.chdir(self.savedcwd)
        shutil.rmtree(self.tempdir)
        sys.argv = self.saved_argv

    def test_correct_configuration(self):
        application = FordoniaApp()
        application.run(dry=True)

    def test_wrong_configuration(self):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_dir = mydir
                '''))
        application = FordoniaApp()
        self.assertRaisesRegex(configparser.Error, 'target_step',
                               application.run)

        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_dir = mydir
                target_step = 60, 0
                nonexistent_option = irrelevant
                '''))
        application = FordoniaApp()
        self.assertRaisesRegex(configparser.Error, 'nonexistent_option',
                               application.run)

        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                base_dir = mydir
                target_step = 60, 80
                '''))
        application = FordoniaApp()
        self.assertRaisesRegex(WrongValueError, 'appropriate time step',
                               application.run)

    def test_execute(self):
        application = FordoniaApp()
        application.read_command_line()
        application.read_configuration()

        # Verify the output files don't exist yet
        self.assertFalse(os.path.exists(self.output_filenames[0]))
        self.assertFalse(os.path.exists(self.output_filenames[1]))

        # Execute
        application.run()

        # Check that it has created two files
        self.assertTrue(os.path.exists(self.output_filenames[0]))
        self.assertTrue(os.path.exists(self.output_filenames[1]))

        # Check that the created time series are correct
        t = Timeseries()
        with open(self.output_filenames[0]) as f:
            t.read_file(f)
        self.assertEqual(len(t), 1)
        self.assertAlmostEqual(t['2014-06-16 16:00'], 114.9, places=5)
        t = Timeseries()
        with open(self.output_filenames[1]) as f:
            t.read_file(f)
        self.assertEqual(len(t), 1)
        self.assertAlmostEqual(t['2014-06-17 16:00'], 50.8167, places=5)

    def test_execute_error_message(self):
        application = FordoniaApp()
        application.read_command_line()
        application.read_configuration()

        # Create a file that doesn't have nominal offset
        with open(self.filenames[0], 'w') as f:
            f.write("Location=23.78743 37.97385 4326\n"
                    "Time_step=10,0\n"
                    "Actual_offset=0,0\n"
                    "\n"
                    "2014-06-16 14:50,14.1,\n"
                    )

        # Execute
        try:
            application.run()
            self.assertTrue(False)
        except Exception as e:
            # Make sure the file name is included in the message
            self.assertTrue(self.filenames[0] in str(e))



from datetime import datetime
import os
import shutil
from six.moves import configparser
import sys
import tempfile
import textwrap
from unittest import TestCase, skipIf

if sys.platform != 'win32':
    import numpy as np
    from osgeo import ogr, gdal
    from pthelma.spatial import BitiaApp, create_ogr_layer_from_timeseries, \
        h_integrate, idw, integrate, WrongValueError
    skip_osgeo = False
    skip_osgeo_message = ''
else:
    skip_osgeo = True
    skip_osgeo_message = 'Not available on Windows'


def add_point_to_layer(layer, x, y, value):
    p = ogr.Geometry(ogr.wkbPoint)
    p.AddPoint(x, y)
    f = ogr.Feature(layer.GetLayerDefn())
    f.SetGeometry(p)
    f.SetField('value', value)
    layer.CreateFeature(f)


@skipIf(skip_osgeo, skip_osgeo_message)
class IdwTestCase(TestCase):

    def setUp(self):
        self.point = ogr.Geometry(ogr.wkbPoint)
        self.point.AddPoint(5.1, 2.5)

        self.data_source = ogr.GetDriverByName('memory').CreateDataSource(
            'tmp')
        self.data_layer = self.data_source.CreateLayer('test')
        self.data_layer.CreateField(ogr.FieldDefn('value', ogr.OFTReal))

    def tearDown(self):
        self.data_layer = None
        self.data_source = None
        self.point = None

    def test_idw_single_point(self):
        add_point_to_layer(self.data_layer, 5.3, 6.4, 42.8)
        self.assertAlmostEqual(idw(self.point, self.data_layer), 42.8)

    def test_idw_three_points(self):
        add_point_to_layer(self.data_layer, 6.4, 7.8, 33.0)
        add_point_to_layer(self.data_layer, 9.5, 7.4, 94.0)
        add_point_to_layer(self.data_layer, 7.1, 4.9, 67.7)
        self.assertAlmostEqual(idw(self.point, self.data_layer),
                               64.090, places=3)
        self.assertAlmostEqual(idw(self.point, self.data_layer, alpha=2.0),
                               64.188, places=3)

    def test_idw_point_with_nan(self):
        add_point_to_layer(self.data_layer, 6.4, 7.8, 33.0)
        add_point_to_layer(self.data_layer, 9.5, 7.4, 94.0)
        add_point_to_layer(self.data_layer, 7.1, 4.9, 67.7)
        add_point_to_layer(self.data_layer, 7.2, 5.4, float('nan'))
        self.assertAlmostEqual(idw(self.point, self.data_layer),
                               64.090, places=3)
        self.assertAlmostEqual(idw(self.point, self.data_layer, alpha=2.0),
                               64.188, places=3)


@skipIf(skip_osgeo, skip_osgeo_message)
class IntegrateTestCase(TestCase):

    def setUp(self):
        # We will test on a 7x15 grid
        self.mask = np.zeros((7, 15), np.int8)
        self.mask[3, 3] = 1
        self.mask[6, 14] = 1
        self.mask[4, 13] = 1
        self.dataset = gdal.GetDriverByName('mem').Create('test', 15, 7, 1,
                                                          gdal.GDT_Byte)
        self.dataset.GetRasterBand(1).WriteArray(self.mask)

        # Prepare the result band
        self.dataset.AddBand(gdal.GDT_Float32)
        self.target_band = self.dataset.GetRasterBand(self.dataset.RasterCount)

        # Our grid represents a 70x150m area, lower-left co-ordinates (0, 0).
        self.dataset.SetGeoTransform((0, 10, 0, 70, 0, -10))

        # Now the data layer, with three points
        self.data_source = ogr.GetDriverByName('memory').CreateDataSource(
            'tmp')
        self.data_layer = self.data_source.CreateLayer('test')
        self.data_layer.CreateField(ogr.FieldDefn('value', ogr.OFTReal))
        add_point_to_layer(self.data_layer,  75.2, 10.7, 37.4)
        add_point_to_layer(self.data_layer, 125.7, 19.0, 24.0)
        add_point_to_layer(self.data_layer,   9.8, 57.1, 95.4)

    def tearDown(self):
        self.data_source = None
        self.raster = None

    def test_integrate_idw(self):
        integrate(self.dataset, self.data_layer, self.target_band, idw)
        result = self.target_band.ReadAsArray()

        # All masked points and only those must be NaN
        # (^ is bitwise xor in Python)
        self.assertTrue((np.isnan(result) ^ (self.mask != 0)).all())

        self.assertAlmostEqual(result[3, 3],  62.971, places=3)
        self.assertAlmostEqual(result[6, 14], 34.838, places=3)
        self.assertAlmostEqual(result[4, 13], 30.737, places=3)


@skipIf(skip_osgeo, skip_osgeo_message)
class CreateOgrLayerFromTimeseriesTestCase(TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

        # Create two time series
        with open(os.path.join(self.tempdir, 'ts1'), 'w') as f:
            f.write(textwrap.dedent("""\
                                    Location=23.78743 37.97385 4326

                                    """))
        with open(os.path.join(self.tempdir, 'ts2'), 'w') as f:
            f.write(textwrap.dedent("""\
                                    Location=24.56789 38.76543 4326

                                    """))

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_create_ogr_layer_from_timeseries(self):
        data_source = ogr.GetDriverByName('memory').CreateDataSource('tmp')
        filenames = [os.path.join(self.tempdir, x) for x in ('ts1', 'ts2')]
        layer = create_ogr_layer_from_timeseries(filenames, 2100, data_source)
        self.assertTrue(layer.GetFeatureCount(), 2)
        # The co-ordinates below are converted from WGS84 to Greek Grid/GGRS87
        ref = [{'x': 481180.63, 'y': 4202647.01,  # 23.78743, 37.97385
                'filename': filenames[0]},
               {'x': 549187.44, 'y': 4290612.25,  # 24.56789, 38.76543
                'filename': filenames[1]},
               ]
        for feature, r in zip(layer, ref):
            self.assertEqual(feature.GetField('filename'), r['filename'])
            self.assertAlmostEqual(feature.GetGeometryRef().GetX(), r['x'], 2)
            self.assertAlmostEqual(feature.GetGeometryRef().GetY(), r['y'], 2)


@skipIf(skip_osgeo, skip_osgeo_message)
class HIntegrateTestCase(TestCase):
    """
    We will test on a 4x3 raster:

                        ABCD
                       1
                       2
                       3

    There will be three stations: two will be exactly at the center of
    gridpoints A3 and B3, and one will be slightly outside the grid (somewhere
    in E2). We assume the bottom left corner of A3 to have co-ordinates (0, 0)
    and the gridpoint to be 10 km across. We also consider C1 to be masked out.
    """

    def create_mask(self):
        mask_array = np.ones((3, 4), np.int8)
        mask_array[0, 2] = 0
        self.mask = gdal.GetDriverByName('mem').Create('mask', 4, 3, 1,
                                                       gdal.GDT_Byte)
        self.mask.SetGeoTransform((0, 10000, 0, 30000, 0, -10000))
        self.mask.GetRasterBand(1).WriteArray(mask_array)

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

        self.filenames = [os.path.join(self.tempdir, x)
                          for x in ('ts1', 'ts2', 'ts3', 'ts4')]
        with open(self.filenames[0], 'w') as f:
            f.write("Location=19.55729 0.04733 2100\n"  # GGRS87=5000 5000
                    "\n"
                    "2014-04-22 12:50,5.03,\n"
                    "2014-04-22 13:00,0.54,\n"
                    "2014-04-22 13:10,6.93,\n")
        with open(self.filenames[1], 'w') as f:
            f.write("Location=19.64689 0.04734 2100\n"  # GGRS87=15000 5000
                    "\n"
                    "2014-04-22 12:50,2.90,\n"
                    "2014-04-22 13:00,2.40,\n"
                    "2014-04-22 13:10,9.16,\n")
        with open(self.filenames[2], 'w') as f:
            f.write("Location=19.88886 0.12857 2100\n"  # GGRS87=42000 14000
                    "\n"
                    "2014-04-22 12:50,9.70,\n"
                    "2014-04-22 13:00,1.84,\n"
                    "2014-04-22 13:10,7.63,\n")
        with open(self.filenames[3], 'w') as f:
             # This station is missing the date required,
             # so it should not be taken into account
            f.write("Location=19.66480 0.15560 2100\n"  # GGRS87=17000 17000
                    "\n"
                    "2014-04-22 12:50,9.70,\n"
                    "2014-04-22 13:10,7.63,\n")

        self.create_mask()
        self.stations = ogr.GetDriverByName('memory').CreateDataSource('tmp')
        self.stations_layer = create_ogr_layer_from_timeseries(
            self.filenames, 2100, self.stations)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_h_integrate(self):
        h_integrate(mask=self.mask,
                    stations_layer=self.stations_layer,
                    date=datetime(2014, 4, 22, 13, 0),
                    output_filename=os.path.join(self.tempdir, 'test.tif'),
                    date_fmt='%Y-%m-%d %H:%M', funct=idw, kwargs={})
        f = gdal.Open(os.path.join(self.tempdir, 'test.tif'))
        result = f.GetRasterBand(1).ReadAsArray()
        expected_result = np.array([[1.5088, 1.6064, np.nan, 1.7237],
                                    [1.3828, 1.6671, 1.7336, 1.7662],
                                    [0.5400, 2.4000, 1.7954, 1.7504]])
        np.testing.assert_almost_equal(result, expected_result, decimal=4)


@skipIf(skip_osgeo, skip_osgeo_message)
class BitiaAppTestCase(TestCase):

    def __init__(self, *args, **kwargs):
        super(BitiaAppTestCase, self).__init__(*args, **kwargs)

        # Python 2.7 compatibility
        try:
            self.assertRaisesRegex
        except AttributeError:
            self.assertRaisesRegex = self.assertRaisesRegexp

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.output_dir = os.path.join(self.tempdir, 'output')
        self.config_file = os.path.join(self.tempdir, 'bitia.conf')
        os.mkdir(self.output_dir)
        self.mask_file = os.path.join(self.tempdir, 'mask.tif')
        self.saved_argv = sys.argv
        sys.argv = ['bitia', '--traceback', self.config_file]

        # Create two time series
        self.filenames = [os.path.join(self.tempdir, x)
                          for x in ('ts1', 'ts2')]
        with open(self.filenames[0], 'w') as f:
            f.write("Location=23.78743 37.97385 4326\n"
                    "Time_step=60,0\n"
                    "\n")
        with open(self.filenames[1], 'w') as f:
            f.write("Location=24.56789 38.76543 4326\n"
                    "Time_step=60,0\n"
                    "\n")

        # Prepare a configuration file (some tests override it)
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                mask = {0.mask_file}
                epsg = 2100
                output_dir = {0.output_dir}
                filename_prefix = rainfall
                files_to_produce = 24
                method = idw
                files = {0.filenames[0]}
                        {0.filenames[1]}
                ''').format(self))

    def tearDown(self):
        shutil.rmtree(self.tempdir)
        sys.argv = self.saved_argv

    def test_correct_configuration(self):
        application = BitiaApp()
        application.run(dry=True)

    def test_wrong_configuration1(self):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                mask = {0.mask_file}
                epsg = 2100
                output_dir = {0.output_dir}
                filename_prefix = rainfall
                method = idw
                files = myfile
                ''').format(self))
        application = BitiaApp()
        self.assertRaisesRegex(configparser.Error, 'files_to_produce',
                               application.run)

    def test_wrong_configuration2(self):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                mask = {0.mask_file}
                epsg = 2100
                output_dir = {0.output_dir}
                filename_prefix = rainfall
                files_to_produce = 24
                method = idw
                files = myfile
                nonexistent_option = irrelevant
                ''').format(self))
        application = BitiaApp()
        self.assertRaisesRegex(configparser.Error, 'nonexistent_option',
                               application.run)

    def test_wrong_configuration_epsg(self):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                mask = {0.mask_file}
                epsg = 81122
                output_dir = {0.output_dir}
                filename_prefix = rainfall
                files_to_produce = 24
                method = idw
                files = myfile
                ''').format(self))
        application = BitiaApp()
        self.assertRaisesRegex(WrongValueError, 'epsg=81122',
                               application.run)

    def test_last_date(self):
        application = BitiaApp()
        application.read_command_line()
        application.read_configuration()
        with open(application.files[0], 'a') as f:
            f.write(textwrap.dedent('''\
                2014-04-30 11:00,18.3,
                2014-04-30 12:00,19.3,
                2014-04-30 13:00,20.4,
                2014-04-30 14:00,21.4,
                '''))
        self.assertEquals(application.last_date, datetime(2014, 4, 30, 14, 0))

    def test_date_fmt(self):
        application = BitiaApp()
        application.read_command_line()
        application.read_configuration()

        # Ten-minute
        with open(application.files[0], 'w') as f:
            f.write('Time_step=10,0\n\n')
        with open(application.files[1], 'w') as f:
            f.write('Time_step=10,0\n\n')
        self.assertEquals(application.date_fmt, '%Y-%m-%d %H:%M')

        # Hourly
        with open(application.files[0], 'w') as f:
            f.write('Time_step=60,0\n\n')
        with open(application.files[1], 'w') as f:
            f.write('Time_step=60,0\n\n')
        self.assertEquals(application.date_fmt, '%Y-%m-%d %H:%M')

        # Daily
        with open(application.files[0], 'w') as f:
            f.write('Time_step=1440,0\n\n')
        with open(application.files[1], 'w') as f:
            f.write('Time_step=1440,0\n\n')
        self.assertEquals(application.date_fmt, '%Y-%m-%d')

        # Monthly
        with open(application.files[0], 'w') as f:
            f.write('Time_step=0,1\n\n')
        with open(application.files[1], 'w') as f:
            f.write('Time_step=0,1\n\n')
        self.assertEquals(application.date_fmt, '%Y-%m')

        # Annual
        with open(application.files[0], 'w') as f:
            f.write('Time_step=0,12\n\n')
        with open(application.files[1], 'w') as f:
            f.write('Time_step=0,12\n\n')
        self.assertEquals(application.date_fmt, '%Y')

        # Inconsistent
        with open(application.files[0], 'w') as f:
            f.write('Time_step=10,0\n\n')
        with open(application.files[1], 'w') as f:
            f.write('Time_step=60,0\n\n')
        self.assertRaises(WrongValueError, lambda: application.date_fmt)

    def get_file_timestamp(self, filename):
        fp = gdal.Open(filename)
        timestamp = fp.GetMetadata()['TIMESTAMP']
        fp = None
        return timestamp

    def create_file_with_timestamp(self, filename, timestamp):
        output = gdal.GetDriverByName('GTiff').Create(filename, 640, 480, 1,
                                                      gdal.GDT_Float32)
        if not output:
            raise IOError('Could not create ' + filename)
        output.SetMetadataItem('TIMESTAMP', timestamp)
        output = None

    def test_rename_existing_files(self):
        application = BitiaApp()
        application.read_command_line()
        application.read_configuration()

        # Annual time step
        for filename in application.files:
            with open(filename, 'w') as f:
                f.write('Time_step=0,12\n\n')

        # Create three files
        prefix = application.config['General']['filename_prefix']
        filename1 = os.path.join(self.output_dir, '{}-0000.tif'.format(prefix))
        filename2 = os.path.join(self.output_dir, '{}-0001.tif'.format(prefix))
        filename3 = os.path.join(self.output_dir, '{}-0002.tif'.format(prefix))
        self.create_file_with_timestamp(filename1, '2014')
        self.create_file_with_timestamp(filename2, '2013')
        self.create_file_with_timestamp(filename3, '2012')

        # Just to make sure we didn't screw up above, check
        self.assertTrue(os.path.exists(filename1))
        self.assertTrue(os.path.exists(filename2))
        self.assertTrue(os.path.exists(filename3))

        # Execute for files_to_produce = 3 and check
        application.config['General']['files_to_produce'] = 3
        application.rename_existing_files('2015')
        self.assertFalse(os.path.exists(filename1))
        self.assertEquals(self.get_file_timestamp(filename2), '2014')
        self.assertEquals(self.get_file_timestamp(filename3), '2013')
        self.assertEquals(len(os.listdir(self.output_dir)), 2)

        # Re-execute; nothing should have changed
        application.rename_existing_files('2015')
        self.assertFalse(os.path.exists(filename1))
        self.assertEquals(self.get_file_timestamp(filename2), '2014')
        self.assertEquals(self.get_file_timestamp(filename3), '2013')
        self.assertEquals(len(os.listdir(self.output_dir)), 2)

    def test_execute(self):
        application = BitiaApp()
        application.read_command_line()
        application.read_configuration()

        # Create some time series data
        with open(application.files[0], 'a') as f:
            f.write(textwrap.dedent('''\
                2014-04-30 11:00,18.3,
                2014-04-30 13:00,20.4,
                2014-04-30 14:00,21.4,
                2014-04-30 15:00,22.4,
                '''))
        with open(application.files[1], 'a') as f:
            f.write(textwrap.dedent('''\
                2014-04-30 11:00,18.3,
                2014-04-30 12:00,19.3,
                2014-04-30 13:00,20.4,
                2014-04-30 14:00,21.4,
                '''))

        # Create a mask
        mask_filename = os.path.join(self.tempdir, 'mask.tif')
        gdal.GetDriverByName('GTiff').Create(mask_filename, 640, 480, 1,
                                             gdal.GDT_Float32)

        # Execute
        application.config['General']['files_to_produce'] = 3
        application.execute()

        # Check that it has created three files
        full_prefix = os.path.join(
            self.output_dir,
            application.config['General']['filename_prefix'])
        self.assertTrue(os.path.exists(full_prefix + '-0000.tif'))
        self.assertTrue(os.path.exists(full_prefix + '-0001.tif'))
        self.assertTrue(os.path.exists(full_prefix + '-0002.tif'))

        # We could check a myriad other things here, but since we've
        # unit-tested lower level functions in detail, the above is reasonably
        # sufficient for us to know that it works.



#!/usr/bin/python
"""
meteologger - utilities to read files of meteorological loggers

Copyright (C) 2005-2011 National Technical University of Athens
Copyright (C) 2005 Antonios Christofides

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

from copy import copy
from datetime import datetime, timedelta
from glob import glob
import math
import os
import re
import struct
import sys
import traceback

from six import StringIO

from pytz import timezone
from simpletail import ropen

from pthelma import enhydris_api
from pthelma.cliapp import CliApp
from pthelma.timeseries import add_months_to_datetime, datetime_from_iso, \
    isoformat_nosecs, Timeseries


class MeteologgerError(Exception):
    pass


class MeteologgerReadError(MeteologgerError):
    pass


class MeteologgerServerError(MeteologgerError):
    pass


class ConfigurationError(MeteologgerError):
    pass


def _next_month(month, year):
    month += 1
    if month > 12:
        month = 1
        year += 1
    return month, year


def _diff_dow(dow1, dow2):
    """
    How many days we need to add to week day dow2 to reach dow1.
    For example, if dow2 is Monday and dow1 is Sunday, the result is 6.
    Monday is 1, Sunday is 7.
    """
    result = dow1 - dow2
    if result < 0:
        result = 7 + result
    return result


def _diff_months(month1, month2):
    result = abs(month2 - month1)
    if result > 6:
        result = 12 - result
    return result


class Datafile(object):
    required_options = ['filename', 'datafile_format', 'datafile_fields']
    optional_options = ['nullstr', 'timezone']

    def __init__(self, base_url, cookies, datafiledict, logger=None):
        self.check_config(datafiledict)
        self.base_url = base_url
        self.cookies = cookies
        self.filename = datafiledict['filename']
        self.datafile_fields = [
            int(x) for x in datafiledict.get('datafile_fields', '').split(',')
            if x != '']
        self.subset_identifiers = datafiledict.get('subset_identifiers', '')
        self.delimiter = datafiledict.get('delimiter', None)
        self.decimal_separator = datafiledict.get('decimal_separator', '')
        self.date_format = datafiledict.get('date_format', '')
        self.nullstr = datafiledict.get('nullstr', '')
        self.nfields_to_ignore = int(datafiledict.get('nfields_to_ignore',
                                     '0'))
        self.timezone = timezone(datafiledict.get('timezone', 'UTC'))
        self.logger = logger
        if not self.logger:
            import logging
            self.logger = logging.getLogger('datafile')
            self.logger.setLevel(logging.WARNING)
            self.logger.addHandler(logging.StreamHandler())

    def check_config(self, datafiledict):
        for option in self.required_options:
            if option not in datafiledict:
                raise ConfigurationError('Option "{}" is required'
                                         .format(option))
        all_options = self.required_options + self.optional_options
        for option in datafiledict:
            if option not in all_options:
                raise ConfigurationError('Unknown option "{}"'.format(option))

    def update_database(self):
        self.logger.info('Processing datafile %s' % (self.filename))
        self.last_timeseries_end_date = None
        self.seq = 0  # sequence number of timeseries
        for self.ts in self.datafile_fields:
            self.seq = self.seq + 1
            if self.ts == 0:
                self.logger.info('Omitting position %d' % (self.seq))
                continue
            self.logger.info('Processing timeseries %d at position %d'
                             % (self.ts, self.seq))
            self._update_timeseries()

    def _get_records_to_append(self, end_date_in_database):
        if (not self.last_timeseries_end_date) or (
                end_date_in_database != self.last_timeseries_end_date):
            self.last_timeseries_end_date = end_date_in_database
            self.logger.info('Reading datafile tail')
            self._get_tail()
            self.logger.info('%d lines in datafile tail' % (len(self.tail)))
            if len(self.tail) > 0:
                self.logger.info('First date in datafile tail: %s' %
                                 (self.tail[0]['date'].isoformat(),))
        ts = Timeseries(self.ts)
        try:
            for line in self.tail:
                v, f = self.extract_value_and_flags(line['line'], self.seq)
                if self.decimal_separator and (self.decimal_separator != '.'):
                    v = v.replace(self.decimal_separator, '.')
                ts[isoformat_nosecs(line['date'])] = (v, f)
        except ValueError as e:
            message = 'parsing error while trying to read values: ' + str(e)
            self.raise_error(line, message)

        return ts

    def _update_timeseries(self):
        end_date_in_database = enhydris_api.get_ts_end_date(
            self.base_url, self.cookies, self.ts)
        self.logger.info('Last date in database: {}'
                         .format(end_date_in_database))
        ts_to_append = self._get_records_to_append(end_date_in_database)
        self.logger.info('Appending %d records' % (len(ts_to_append)))
        if len(ts_to_append):
            self.logger.info('First appended record: %s' %
                            (ts_to_append.items()[0][0].isoformat()))
            self.logger.info('Last appended record:  %s' %
                            (ts_to_append.items()[-1][0].isoformat()))
        self._append_records_to_database(ts_to_append)

    def _append_records_to_database(self, ts_to_append):
        fp = StringIO()
        ts_to_append.write(fp)
        r = enhydris_api.post_tsdata(self.base_url, self.cookies, ts_to_append)
        self.logger.info(
            'Successfully appended {0} records'.format(r))

    def _get_tail(self):
        "Read the part of the datafile after last_timeseries_end_date"
        self.tail = []
        prev_date = ''
        with ropen(self.filename) as xr:
            for line in xr:
                self.logger.debug(line)
                if line.strip() and self.subset_identifiers_match(line):
                    date = self.extract_date(line).replace(second=0)
                    date = self._fix_dst(date)
                    if date == prev_date:
                        w = 'WARNING: Omitting line with repeated date ' + str(
                            date)
                        self.logger.warning(w)
                        continue
                    prev_date = date
                    self.logger.debug('Date: %s' % (date.isoformat()))
                    if date <= self.last_timeseries_end_date:
                        break
                    self.tail.append({'date': date, 'line': line})
        self.tail.reverse()

    def _fix_dst(self, adatetime, now=None):
        """Remove any DST from a date.

           Determine if a date contains DST. If it does, remove the
           extra hour. Returns the fixed date.

           The "now" argument is only used for testing. Normally the method
           assigns the current datetime to it. However, when testing, we
           sometimes want to pretend the time is different.
        """
        result = adatetime
        if self.timezone.zone != 'UTC':
            if not now:
                now = datetime.now(self.timezone)
            now_naive = now.replace(tzinfo=None)
            is_dst = bool(now.dst()) and (
                abs(adatetime - now_naive) < timedelta(hours=24))
            result -= self.timezone.dst(adatetime, is_dst=is_dst)
        return result

    def subset_identifiers_match(self, line):
        "Returns true if subset identifier of line matches specified"
        return True

    def extract_date(self, line):
        raise Exception(
            "Internal error: datafile.extract_date is an abstract function")

    def extract_value_and_flags(self, line, seq):
        raise Exception("Internal error: datafile.extract_value_and_flags "
                        "is an abstract function")

    def raise_error(self, line, msg):
        errmessage = '%s: "%s": %s' % (self.filename, line, msg)
        self.logger.error("Error while parsing, message: %s" % errmessage)
        raise MeteologgerReadError(errmessage)


class Datafile_deltacom(Datafile):
    deltacom_flags = {'#': 'LOGOVERRUN',
                      '$': 'LOGNOISY',
                      '%': 'LOGOUTSIDE',
                      '&': 'LOGRANGE'}

    def extract_date(self, line):
        try:
            return datetime_from_iso(line)
        except ValueError:
            self.raise_error(line, 'parse error or invalid date')

    def extract_value_and_flags(self, line, seq):
        flags = ''
        item = line.split()[seq].strip()
        if item[-1] in self.deltacom_flags.keys():
            flags = self.deltacom_flags[item[-1]]
            item = item[:-1]
        if self.nullstr:
            if item == self.nullstr:
                item = float('NaN')
        return (item, flags)


class Datafile_pc208w(Datafile):
    required_options = Datafile.required_options + [
        'subset_identifiers']

    def extract_date(self, line):
        try:
            items = line.split(',')
            year = int(items[2])
            yday = int(items[3])
            hour = int(items[4]) / 100
            minute = int(items[4]) % 100
            if hour == 24:
                hour = 0
                yday = yday + 1
            return datetime(year, 1, 1, hour, minute) + timedelta(yday - 1)
        except StandardError:
            self.raise_error(line, 'parse error or invalid date')

    def extract_value_and_flags(self, line, seq):
        try:
            item = line.split(',')[seq + 4].strip()
        except IndexError:
            raise ValueError()
        if self.nullstr:
            if item == self.nullstr:
                item = float('NaN')
        return (item, '')

    def subset_identifiers_match(self, line):
        si = line.split(',')[0].strip()
        return si == self.subset_identifiers


class Datafile_CR1000(Datafile):
    required_options = Datafile.required_options + [
        'subset_identifiers']

    def extract_date(self, line):
        try:
            datestr = line.split(',')[0].strip('"')
            return datetime_from_iso(datestr[:16])
        except StandardError:
            self.raise_error(line, 'parse error or invalid date')

    def extract_value_and_flags(self, line, seq):
        return (line.split(',')[seq + 3].strip(), '')

    def subset_identifiers_match(self, line):
        si = line.split(',')[2].strip()
        return si == self.subset_identifiers


class Datafile_simple(Datafile):
    optional_options = Datafile.optional_options + [
        'nfields_to_ignore', 'delimiter', 'date_format']

    def __init__(self, base_url, cookies, datafiledict, logger=None):
        super(Datafile_simple, self).__init__(base_url, cookies, datafiledict,
                                              logger)
        self.__separate_time = False

    def extract_date(self, line):
        try:
            items = line.split(self.delimiter)
            datestr = items[self.nfields_to_ignore].strip('"')
            self.__separate_time = False
            if len(datestr) <= 10:
                datestr += ' ' + items[self.nfields_to_ignore + 1].strip('"')
                self.__separate_time = True
            if self.date_format:
                result = datetime.strptime(datestr, self.date_format).replace(
                    second=0)
            else:
                result = datetime_from_iso(datestr[:16])
            return result
        except ValueError as e:
            self.raise_error(line.strip(), "invalid date '{0}': {1}".format(
                datestr, str(e)))

    def extract_value_and_flags(self, line, seq):
        index = self.nfields_to_ignore + seq + (
            1 if self.__separate_time else 0)
        value = line.split(self.delimiter)[index].strip().strip('"').strip()
        if self.nullstr and value == self.nullstr:
            value = float('NaN')
        return (value, '')


class Datafile_lastem(Datafile):
    required_options = Datafile.required_options + [
        'subset_identifiers']
    optional_options = Datafile.optional_options + [
        'delimiter', 'decimal_separator', 'date_format']

    def extract_date(self, line):
        try:
            date = line.split(self.delimiter)[3]
            return datetime.strptime(date, self.date_format)
        except StandardError:
            self.raise_error(line, 'parse error or invalid date')

    def extract_value_and_flags(self, line, seq):
        value = line.split(self.delimiter)[seq + 3]
        if self.nullstr:
            if value == self.nullstr:
                value = float('NaN')
        if not math.isnan(value):
            value = value.replace(self.decimal_separator, '.')
        return (value, '')

    def subset_identifiers_match(self, line):
        si = [x.strip() for x in line.split(self.delimiter)[0:3]]
        si1 = [x.strip() for x in self.subset_identifiers.split(',')]
        return si == si1


class Datafile_wdat5(Datafile):
    wdat_record_format = [
        '<b dataType',
        '<b archiveInterval',
        '<b iconFlags',
        '<b moreFlags',
        '<h packedTime',
        '<h outsideTemp',
        '<h hiOutsideTemp',
        '<h lowOutsideTemp',
        '<h insideTemp',
        '<h barometer',
        '<h outsideHum',
        '<h insideHum',
        '<H rain',
        '<h hiRainRate',
        '<h windSpeed',
        '<h hiWindSpeed',
        '<b windDirection',
        '<b hiWindDirection',
        '<h numWindSamples',
        '<h solarRad',
        '<h hiSolarRad',
        '<B UV',
        '<B hiUV',
        '<b leafTemp1', '<b leafTemp2',
        '<b leafTemp3', '<b leafTemp4',
        '<h extraRad',
        '<h newSensors1', '<h newSensors2', '<h newSensors3',
        '<h newSensors4', '<h newSensors5', '<h newSensors6',
        '<b forecast',
        '<B ET',
        '<b soilTemp1', '<b soilTemp2', '<b soilTemp3',
        '<b soilTemp4', '<b soilTemp5', '<b soilTemp6',
        '<b soilMoisture1', '<b soilMoisture2', '<b soilMoisture3',
        '<b soilMoisture4', '<b soilMoisture5', '<b soilMoisture6',
        '<b leafWetness1', '<b leafWetness2',
        '<b leafWetness3', '<b leafWetness4',
        '<b extraTemp1', '<b extraTemp2', '<b extraTemp3',
        '<b extraTemp4', '<b extraTemp5', '<b extraTemp6',
        '<b extraTemp7',
        '<b extraHum1', '<b extraHum2', '<b extraHum3',
        '<b extraHum4', '<b extraHum5', '<b extraHum6',
        '<b extraHum7',
    ]
    required_options = [x for x in Datafile.required_options
                        if x != 'datafile_fields']
    variables_labels = [x.split()[1].lower() for x in wdat_record_format[5:]]
    optional_options = variables_labels + ['timezone']

    def __init__(self, base_url, cookies, datafiledict, logger=None):
        super(Datafile_wdat5, self).__init__(base_url, cookies, datafiledict,
                                             logger)

        self.variables = {}
        for label in self.variables_labels:
            self.variables[label] = datafiledict.get(label, None)

        unit_parameters = {
            'temperature_unit': ('C', 'F'),
            'rain_unit': ('mm', 'inch'),
            'wind_speed_unit': ('m/s', 'mph'),
            'pressure_unit': ('hPa', 'inch Hg'),
            'matric_potential_unit': ('centibar', 'cm'),
        }
        for p in unit_parameters:
            self.__dict__[p] = datafiledict.get(p, unit_parameters[p][0])
            if self.__dict__[p] not in unit_parameters[p]:
                raise ConfigurationError("{0} must be one of {1}".format(
                    p, ', '.join(unit_parameters[p])))

    def update_database(self):
        self.logger.info('Processing data directory %s' % (self.filename))
        self.last_timeseries_end_date = None
        for self.seq, label in enumerate(self.variables_labels, start=1):
            self.ts = self.variables.get(label, None)
            if not self.ts:
                self.logger.debug('Omitting {}'.format(label))
                continue
            self.logger.info('Processing timeseries {} as {}'.format(self.ts,
                                                                     label))
            self._update_timeseries()

    def _get_tail(self):
        "Read the part of the data after last_timeseries_end_date"
        self.tail = []
        date = self.last_timeseries_end_date
        saveddir = os.getcwd()
        try:
            os.chdir(self.filename)

            # We assume that the first possible file we need to read is the
            # previous than what it seems initially, because of the possibility
            # of a DST offset interfering (I think that actually a negative DST
            # offset would be needed for this to be a problem - but let's be
            # safe and not sorry) (note that if the date is 0001-01-01 it
            # means that the time series in the database is empty and we'll
            # upload it in its entirety anyway, so no need to do the trick
            # [which would fail because dates can't be less than that]).
            a = add_months_to_datetime(date, -1) if date.year > 1 else date
            first_file = '{0.year:04}-{0.month:02}.wlk'.format(a)

            filename_regexp = re.compile(r'\d{4}-\d{2}.wlk$')
            data_files = [x for x in glob('*.wlk')
                          if filename_regexp.match(x) and x >= first_file]
            data_files.sort()
            for current_file in data_files:
                self.tail.extend(self._get_tail_part(date, current_file))
        finally:
            os.chdir(saveddir)

    def _get_tail_part(self, last_date, filename):
        """
        Read a single wdat5 file.

        Reads the single wdat5 file "filename" for records with
        date>last_date, and returns a list of records in space-delimited
        format; iso datetime first, values after.
        """
        year, month = [int(x) for x
                       in os.path.split(filename)[1].split('.')[0].split('-')]
        result = []
        with open(filename, 'rb') as f:
            header = f.read(212)
            if header[:6] != b'WDAT5.':
                raise MeteologgerReadError('File {0} does not appear to be '
                                           'a WDAT 5.x file'.format(filename))
            for day in range(1, 32):
                day_index = header[20 + (day * 6):20 + (day * 6) + 6]
                records_in_day = struct.unpack('<h', day_index[:2])[0]
                start_pos = struct.unpack('<l', day_index[2:])[0]
                for r in range(records_in_day):
                    f.seek(212 + ((start_pos + r) * 88))
                    record = f.read(88)
                    if record[0] != b'\x01'[0]:
                        continue
                    decoded_record = self._decode_wdat_record(record)
                    date = datetime(year=year, month=month, day=day) + \
                        timedelta(minutes=decoded_record['packedTime'])
                    date = self._fix_dst(date)
                    if date <= last_date:
                        continue
                    result.append({'date': date,
                                   'line': self._convert_wdat_record(
                                       date, decoded_record)})
        return result

    def _decode_wdat_record(self, record):
        "Decode bytes into a dictionary"
        result = {}
        offset = 0
        for item in self.wdat_record_format:
            fmt, name = item.split()
            result[name] = struct.unpack_from(fmt, record, offset)[0]
            offset += struct.calcsize(fmt)
        return result

    def _convert_wdat_record(self, date, decoded_record):
        "Return a space-delimited string based on the decoded record"
        r = copy(decoded_record)

        # Temperature
        for x in ['outsideTemp', 'hiOutsideTemp', 'lowOutsideTemp',
                  'insideTemp']:
            r[x] = (r[x] / 10.0 if self.temperature_unit == 'F' else
                    ((r[x] / 10.0) - 32) * 5 / 9.0)

        # Pressure
        r['barometer'] = (r['barometer'] / 1000.0
                          if self.pressure_unit == 'inch Hg'
                          else r['barometer'] / 1000.0 * 25.4 * 1.33322387415)

        # Humidity
        for x in ['outsideHum', 'insideHum']:
            r[x] = r[x] / 10.0

        # Rain
        rain_collector_type = r['rain'] & 0xF000
        rain_clicks = r['rain'] & 0x0FFF
        depth_per_click = {
            0x0000: 0.1 * 25.4,
            0x1000: 0.01 * 25.4,
            0x2000: 0.2,
            0x3000: 1.0,
            0x6000: 0.1,
        }[rain_collector_type]
        depth = depth_per_click * rain_clicks
        r['rain'] = depth / 25.4 if self.rain_unit == 'inch' else depth
        rate = r['hiRainRate'] * depth_per_click
        r['hiRainRate'] = rate / 25.4 if self.rain_unit == 'inch' else rate

        # Wind speed
        convert_wind_speed = lambda x: (x / 10.0
                                        if self.wind_speed_unit == 'mph'
                                        else x / 10.0 * 1609.344 / 3600)
        r['windSpeed'] = convert_wind_speed(r['windSpeed'])
        r['hiWindSpeed'] = convert_wind_speed(r['hiWindSpeed'])

        # Wind direction
        for x in ['windDirection', 'hiWindDirection']:
            r[x] = r[x] / 16.0 * 360 if r[x] >= 0 else 'NaN'

        # UV index
        r['UV'] = r['UV'] / 10.0
        r['hiUV'] = r['hiUV'] / 10.0

        # Evapotranspiration
        r['ET'] = (r['ET'] / 1000.0 if self.rain_unit == 'inch'
                   else r['ET'] / 1000.0 * 25.4)

        # Matric potential
        for i in range(1, 7):
            varname = 'soilMoisture' + str(i)
            r[varname] = (r[varname]
                          if self.matric_potential_unit == 'centibar'
                          else r[varname] / 9.80638)

        # extraTemp etc.
        for x in ['extraTemp1', 'extraTemp2', 'extraTemp3', 'extraTemp4',
                  'extraTemp5', 'extraTemp6', 'extraTemp7', 'soilTemp1',
                  'soilTemp2', 'soilTemp3', 'soilTemp4', 'soilTemp5',
                  'soilTemp6', 'leafTemp1', 'leafTemp2', 'leafTemp3',
                  'leafTemp4']:
            r[x] = (r[x] - 90 if self.temperature_unit == 'F' else
                    ((r[x] - 90) - 32) * 5 / 9.0)

        result = "{r[outsideTemp]} {r[hiOutsideTemp]} " \
                 "{r[lowOutsideTemp]} {r[insideTemp]} {r[barometer]} " \
                 "{r[outsideHum]} {r[insideHum]} {r[rain]} {r[hiRainRate]} " \
                 "{r[windSpeed]} {r[hiWindSpeed]} " \
                 "{r[windDirection]} {r[hiWindDirection]} " \
                 "{r[numWindSamples]} {r[solarRad]} {r[hiSolarRad]} {r[UV]} " \
                 "{r[hiUV]} " \
                 "{r[leafTemp1]} {r[leafTemp2]} " \
                 "{r[leafTemp3]} {r[leafTemp4]} " \
                 "{r[extraRad]} " \
                 "{r[newSensors1]} {r[newSensors2]} {r[newSensors3]} " \
                 "{r[newSensors4]} {r[newSensors5]} {r[newSensors6]} " \
                 "{r[forecast]} {r[ET]} " \
                 "{r[soilTemp1]} {r[soilTemp2]} {r[soilTemp3]} " \
                 "{r[soilTemp4]} {r[soilTemp5]} {r[soilTemp6]} " \
                 "{r[soilMoisture1]} {r[soilMoisture2]} {r[soilMoisture3]} " \
                 "{r[soilMoisture4]} {r[soilMoisture5]} {r[soilMoisture6]} " \
                 "{r[leafWetness1]} {r[leafWetness2]} " \
                 "{r[leafWetness3]} {r[leafWetness4]} " \
                 "{r[extraTemp1]} {r[extraTemp2]} {r[extraTemp3]} " \
                 "{r[extraTemp4]} {r[extraTemp5]} {r[extraTemp6]} " \
                 "{r[extraTemp7]} " \
                 "{r[extraHum1]} {r[extraHum2]} {r[extraHum3]} " \
                 "{r[extraHum4]} {r[extraHum5]} {r[extraHum6]} " \
                 "{r[extraHum7]} ".format(r=r)
        return result

    def extract_value_and_flags(self, line, seq):
        result = line.split()[seq - 1].strip()
        return (result, '')


class Datafile_odbc(Datafile_simple):
    required_options = Datafile.required_options + [
        'table', 'date_sql', 'data_columns']
    optional_options = Datafile.optional_options + ['date_format',
                                                    'decimal_separator']

    def __init__(self, base_url, cookies, datafiledict, logger=None):
        super(Datafile_odbc, self).__init__(base_url, cookies,
                                            datafiledict, logger)
        self.table = datafiledict.get('table', '')
        self.date_sql = datafiledict.get('date_sql', '')
        self.data_columns = datafiledict.get('data_columns', '').split(',')
        self.delimiter = ';'

    def _get_tail(self):
        "Read the part of the datafile after last_timeseries_end_date"
        try:
            import pyodbc
        except ImportError:
            self.logger.error('Install pyodbc to use odbc format')
            raise
        sql = """SELECT {} + ';' + {} FROM "{}" ORDER BY -id""".format(
            self.date_sql,
            " + ';' + ".join(['"{}"'.format(x) for x in self.data_columns]),
            self.table)
        self.tail = []
        connection = pyodbc.connect(self.filename)
        cursor = connection.cursor()
        cursor.execute(sql)
        for row in cursor:  # Iterable cursor is a pyodbc feature
            line = row[0]  # Our SQL returns a single string
            self.logger.debug(line)
            date = self.extract_date(line).replace(second=0)
            date = self._fix_dst(date)
            self.logger.debug('Date: %s' % (date.isoformat()))
            if date <= self.last_timeseries_end_date:
                break
            self.tail.append({'date': date, 'line': line})
        self.tail.reverse()


class LoggertodbApp(CliApp):
    name = 'loggertodb'
    description = 'Insert meteorological logger data to Enhydris'
    config_file_options = {'General': {'base_url': None,
                                       'user':     None,
                                       'password': None,
                                       },
                           'other':   'nocheck',
                           }

    def read_configuration(self):
        super(LoggertodbApp, self).read_configuration()
        if not self.config['General']['base_url'].endswith('/'):
            self.config['General']['base_url'] += '/'

    def execute(self):
        cookies = enhydris_api.login(self.config['General']['base_url'],
                                     self.config['General']['user'],
                                     self.config['General']['password'])
        for section in self.config:
            if section == 'General':
                continue
            datafileclass = eval('Datafile_{}'.format(
                self.config[section]['datafile_format']))
            adatafile = datafileclass(self.config['General']['base_url'],
                                      cookies,
                                      self.config[section],
                                      self.logger)
            try:
                adatafile.update_database()
            except MeteologgerError as e:
                msg = 'Error while processing item {0}: {1}'.format(section,
                                                                    str(e))
                sys.stderr.write(msg + '\n')
                self.logger.error(msg)
                self.logger.debug(traceback.format_exc())



"""
In order to run these tests, you must specify the
PTHELMA_TEST_ENHYDRIS_API variable to contain a json-formatted string
of parameters, like this:

    { "base_url": "http://localhost:8001/",
      "user": "admin",
      "password": "topsecret",
      "station_id": 1334,
      "variable_id": 1,
      "unit_of_measurement_id": 1,
      "time_zone_id": 1,
      "owner_id": 5
    }

The parameters station_id, variable_id, unit_of_measurement_id,
time_zone_id and owner_id are used when test timeseries or stations are
created.

Note that some other testing modules may also use
PTHELMA_TEST_ENHYDRIS_API, but might require additional parameters.

Don't use a production database for that; some testing functionality
may write to the database. Although things are normally cleaned up
(e.g. test timeseries created are deleted), id serial numbers will be
affected and things might not be cleaned up if there is an error.
"""

import json
import os
import textwrap
from unittest import TestCase, skipUnless

from six import StringIO

import requests

from pthelma import enhydris_api
from pthelma.timeseries import Timeseries


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            'set PTHELMA_TEST_ENHYDRIS_API')
class LoginTestCase(TestCase):

    def test_login(self):
        v = json.loads(os.getenv('PTHELMA_TEST_ENHYDRIS_API'))
        base_url, user, password = v['base_url'], v['user'], v['password']

        # Verify we are logged out
        r = requests.get(base_url)
        self.assertEqual(r.status_code, 200)
        self.assertTrue('Login' in r.text)
        self.assertFalse('Logout' in r.text)

        # Now login and verify we're logged on
        cookies = enhydris_api.login(base_url, user, password)
        r = requests.get(base_url, cookies=cookies)
        self.assertEqual(r.status_code, 200)
        self.assertFalse('Login' in r.text)
        self.assertTrue('Logout' in r.text)


def get_after_blank_line(s):
    r"""
    This helper function returns all content of s that follows the
    first empty line; it also removes \r from line endings.
    """
    lines = s.splitlines(True)
    i = 0
    while True:
        if not lines[i].strip():
            break
        i += 1
    lines = lines[i + 1:]
    return ''.join([x.replace('\r', '') for x in lines])


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            'set PTHELMA_TEST_ENHYDRIS_API')
class PostTsDataTestCase(TestCase):
    test_timeseries = textwrap.dedent("""\
                                      2014-01-01 08:00,11,
                                      2014-01-02 08:00,12,
                                      2014-01-03 08:00,13,
                                      2014-01-04 08:00,14,
                                      2014-01-05 08:00,15,
                                      """)
    test_timeseries_top = ''.join(test_timeseries.splitlines(True)[:-1])
    test_timeseries_bottom = ''.join(test_timeseries.splitlines(True)[-1])

    def test_post_tsdata(self):
        v = json.loads(os.getenv('PTHELMA_TEST_ENHYDRIS_API'))
        cookies = enhydris_api.login(v['base_url'], v['user'], v['password'])

        # Create a time series in the database
        j = {
            'gentity': v['station_id'],
            'variable': v['variable_id'],
            'unit_of_measurement': v['unit_of_measurement_id'],
            'time_zone': v['time_zone_id'],
        }
        ts_id = enhydris_api.post_model(v['base_url'], cookies, 'Timeseries',
                                        j)

        # Now upload some data
        ts = Timeseries(ts_id)
        ts.read(StringIO(self.test_timeseries_top))
        enhydris_api.post_tsdata(v['base_url'], cookies, ts)

        # Read and check the time series
        url = enhydris_api.urljoin(v['base_url'],
                                   'timeseries/d/{}/download/'.format(ts.id))
        r = requests.get(url, cookies=cookies)
        r.raise_for_status()
        self.assertEqual(get_after_blank_line(r.text),
                         self.test_timeseries_top)

        # Upload more data
        ts = Timeseries(ts_id)
        ts.read(StringIO(self.test_timeseries_bottom))
        enhydris_api.post_tsdata(v['base_url'], cookies, ts)

        # Read and check the time series
        url = enhydris_api.urljoin(v['base_url'],
                                   'timeseries/d/{}/download/'.format(ts.id))
        r = requests.get(url, cookies=cookies)
        r.raise_for_status()
        self.assertEqual(get_after_blank_line(r.text),
                         self.test_timeseries)


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            'set PTHELMA_TEST_ENHYDRIS_API')
class GetTsEndDateTestCase(TestCase):
    test_timeseries = textwrap.dedent("""\
                                      2014-01-01 08:00,11,
                                      2014-01-02 08:00,12,
                                      2014-01-03 08:00,13,
                                      2014-01-04 08:00,14,
                                      2014-01-05 08:00,15,
                                      """)

    def test_get_ts_end_date(self):
        v = json.loads(os.getenv('PTHELMA_TEST_ENHYDRIS_API'))
        cookies = enhydris_api.login(v['base_url'], v['user'], v['password'])

        # Create a time series in the database
        j = {
            'gentity': v['station_id'],
            'variable': v['variable_id'],
            'unit_of_measurement': v['unit_of_measurement_id'],
            'time_zone': v['time_zone_id'],
        }
        ts_id = enhydris_api.post_model(v['base_url'], cookies, 'Timeseries',
                                        j)

        # Get its last date while it has no data
        date = enhydris_api.get_ts_end_date(v['base_url'], cookies, ts_id)
        self.assertEqual(date.isoformat(), '0001-01-01T00:00:00')

        # Now upload some data
        ts = Timeseries(ts_id)
        ts.read(StringIO(self.test_timeseries))
        enhydris_api.post_tsdata(v['base_url'], cookies, ts)

        # Get its last date
        date = enhydris_api.get_ts_end_date(v['base_url'], cookies, ts_id)
        self.assertEqual(date.isoformat(), '2014-01-05T08:00:00')

        # Get the last date of a nonexistent time series
        self.assertRaises(requests.HTTPError, enhydris_api.get_ts_end_date,
                          v['base_url'], cookies, ts_id + 1)



# -*- coding: utf8 -*-
import json
import os
import shutil
import sys
import tempfile
import textwrap
from unittest import TestCase, skipUnless

from six import StringIO
from six.moves import configparser

from pthelma import enhydris_api
from pthelma.enhydris_cache import PondApp, TimeseriesCache
from pthelma.timeseries import Timeseries


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            'set PTHELMA_TEST_ENHYDRIS_API')
class TimeseriesCacheTestCase(TestCase):
    test_timeseries1 = textwrap.dedent("""\
                                   2014-01-01 08:00,11,
                                   2014-01-02 08:00,12,
                                   2014-01-03 08:00,13,
                                   2014-01-04 08:00,14,
                                   2014-01-05 08:00,15,
                                   """)
    test_timeseries2 = textwrap.dedent("""\
                                   2014-07-01 08:00,9.11,
                                   2014-07-02 08:00,9.12,
                                   2014-07-03 08:00,9.13,
                                   2014-07-04 08:00,9.14,
                                   2014-07-05 08:00,9.15,
                                   """)
    timeseries1_top = ''.join(test_timeseries1.splitlines(True)[:-1])
    timeseries2_top = ''.join(test_timeseries2.splitlines(True)[:-1])
    timeseries1_bottom = test_timeseries1.splitlines(True)[-1]
    timeseries2_bottom = test_timeseries2.splitlines(True)[-1]

    def setUp(self):
        self.parms = json.loads(os.getenv('PTHELMA_TEST_ENHYDRIS_API'))
        self.cookies = enhydris_api.login(self.parms['base_url'],
                                          self.parms['user'],
                                          self.parms['password'])

        # Create two time series
        j = {
            'gentity': self.parms['station_id'],
            'variable': self.parms['variable_id'],
            'unit_of_measurement': self.parms['unit_of_measurement_id'],
            'time_zone': self.parms['time_zone_id'],
            'time_step': 3,
            'actual_offset_minutes': 0,
            'actual_offset_months': 0,
            'remarks': 'Très importante',
        }
        self.ts1_id = enhydris_api.post_model(
            self.parms['base_url'], self.cookies, 'Timeseries', j)
        self.ts2_id = enhydris_api.post_model(
            self.parms['base_url'], self.cookies, 'Timeseries', j)
        assert self.ts1_id != self.ts2_id

        # Add some data (all but the last record) to the database
        ts = Timeseries(self.ts1_id)
        ts.read(StringIO(self.timeseries1_top))
        enhydris_api.post_tsdata(self.parms['base_url'], self.cookies, ts)
        ts = Timeseries(self.ts2_id)
        ts.read(StringIO(self.timeseries2_top))
        enhydris_api.post_tsdata(self.parms['base_url'], self.cookies, ts)

        # Temporary directory for cache files
        self.tempdir = tempfile.mkdtemp()
        self.savedcwd = os.getcwd()
        os.chdir(self.tempdir)

    def tearDown(self):
        os.chdir(self.savedcwd)
        shutil.rmtree(self.tempdir)

    def test_update(self):
        self.parms = json.loads(os.getenv('PTHELMA_TEST_ENHYDRIS_API'))
        timeseries_group = [{'base_url': self.parms['base_url'],
                             'id': self.ts1_id,
                             'user': self.parms['user'],
                             'password': self.parms['password'],
                             'file': 'file1',
                             },
                            {'base_url': self.parms['base_url'],
                             'id': self.ts2_id,
                             'user': self.parms['user'],
                             'password': self.parms['password'],
                             'file': 'file2',
                             },
                            ]
        # Cache the two timeseries
        cache = TimeseriesCache(timeseries_group)
        cache.update()

        # Check that the cached stuff is what it should be
        with open('file1') as f:
            ts = Timeseries()
            ts.read_file(f)
            self.assertEqual(ts.time_step.length_minutes, 1440)
            self.assertEqual(ts.time_step.length_months, 0)
            c = StringIO()
            ts.write(c)
            self.assertEqual(c.getvalue().replace('\r', ''),
                             self.timeseries1_top)
        with open('file2') as f:
            ts = Timeseries()
            ts.read_file(f)
            self.assertEqual(ts.time_step.length_minutes, 1440)
            self.assertEqual(ts.time_step.length_months, 0)
            c = StringIO()
            ts.write(c)
            self.assertEqual(c.getvalue().replace('\r', ''),
                             self.timeseries2_top)

        # Append a record to the database for each timeseries
        ts = Timeseries(self.ts1_id)
        ts.read(StringIO(self.timeseries1_bottom))
        enhydris_api.post_tsdata(self.parms['base_url'], self.cookies, ts)
        ts = Timeseries(self.ts2_id)
        ts.read(StringIO(self.timeseries2_bottom))
        enhydris_api.post_tsdata(self.parms['base_url'], self.cookies, ts)

        # Update the cache
        cache.update()

        # Check that the cached stuff is what it should be
        with open('file1') as f:
            ts = Timeseries()
            ts.read_file(f)
            self.assertEqual(ts.time_step.length_minutes, 1440)
            self.assertEqual(ts.time_step.length_months, 0)
            c = StringIO()
            ts.write(c)
            self.assertEqual(c.getvalue().replace('\r', ''),
                             self.test_timeseries1)
        with open('file2') as f:
            ts = Timeseries()
            ts.read_file(f)
            self.assertEqual(ts.time_step.length_minutes, 1440)
            self.assertEqual(ts.time_step.length_months, 0)
            c = StringIO()
            ts.write(c)
            self.assertEqual(c.getvalue().replace('\r', ''),
                             self.test_timeseries2)


@skipUnless(os.getenv('PTHELMA_TEST_ENHYDRIS_API'),
            'set PTHELMA_TEST_ENHYDRIS_API')
class PondAppTestCase(TestCase):

    def __init__(self, *args, **kwargs):
        super(PondAppTestCase, self).__init__(*args, **kwargs)

        # Python 2.7 compatibility
        try:
            self.assertRaisesRegex
        except AttributeError:
            self.assertRaisesRegex = self.assertRaisesRegexp

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config_file = os.path.join(self.tempdir, 'pond.conf')
        self.saved_argv = sys.argv
        sys.argv = ['pond', '--traceback', self.config_file]
        self.savedcwd = os.getcwd()

        # Create two stations, each one with a time series
        self.parms = json.loads(os.getenv('PTHELMA_TEST_ENHYDRIS_API'))
        self.cookies = enhydris_api.login(self.parms['base_url'],
                                          self.parms['user'],
                                          self.parms['password'])
        self.station1_id = enhydris_api.post_model(
            self.parms['base_url'], self.cookies, 'Station',
            {'name': 'station1',
             'srid': 4326,
             'point': 'POINT (23.78743 37.97385)',
             'copyright_holder': 'Joe User',
             'copyright_years': '2014',
             'stype': 1,
             'owner': self.parms['owner_id'],
             })
        self.timeseries1_id = enhydris_api.post_model(
            self.parms['base_url'], self.cookies, 'Timeseries',
            {'gentity': self.station1_id,
             'variable': self.parms['variable_id'],
             'unit_of_measurement': self.parms['unit_of_measurement_id'],
             'time_zone': self.parms['time_zone_id']})
        self.station2_id = enhydris_api.post_model(
            self.parms['base_url'], self.cookies, 'Station',
            {'name': 'station1',
             'srid': 4326,
             'point': 'POINT (24.56789 38.76543)',
             'copyright_holder': 'Joe User',
             'copyright_years': '2014',
             'stype': 1,
             'owner': self.parms['owner_id'],
             })
        self.timeseries2_id = enhydris_api.post_model(
            self.parms['base_url'], self.cookies, 'Timeseries',
            {'gentity': self.station2_id,
             'variable': self.parms['variable_id'],
             'unit_of_measurement': self.parms['unit_of_measurement_id'],
             'time_zone': self.parms['time_zone_id']})

        # Prepare a configuration file (some tests override it)
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                cache_dir = {self.tempdir}

                [timeseries1]
                base_url = {base_url}
                id = {self.timeseries1_id}
                file = file1

                [timeseries2]
                base_url = {base_url}
                id = {self.timeseries2_id}
                file = file2
                ''').format(self=self, base_url=self.parms['base_url']))

    def tearDown(self):
        os.chdir(self.savedcwd)
        shutil.rmtree(self.tempdir)
        sys.argv = self.saved_argv

    def test_correct_configuration(self):
        application = PondApp()
        application.run(dry=True)

    def test_wrong_configuration1(self):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                cache_dir = {self.tempdir}
                nonexistent_option = irrelevant
                ''').format(self=self))
        application = PondApp()
        self.assertRaisesRegex(configparser.Error, 'nonexistent_option',
                               application.run)

    def test_wrong_configuration2(self):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                cache_dir = {self.tempdir}

                [timeseries1]
                id = 5432
                file = file1
                ''').format(self=self))
        application = PondApp()
        self.assertRaisesRegex(configparser.Error, 'base_url',
                               application.run)

    def test_wrong_configuration3(self):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent('''\
                [General]
                cache_dir = {self.tempdir}

                [timeseries1]
                base_url = wrongproto://irrelevant.com/
                id = non-numeric
                file = file1
                ''').format(self=self))
        application = PondApp()
        self.assertRaisesRegex(configparser.Error, 'id',
                               application.run)

    def test_execute(self):
        application = PondApp()
        application.read_command_line()
        application.read_configuration()

        # Check that the two files don't exist yet
        self.assertFalse(os.path.exists(os.path.join(self.tempdir, 'file1')))
        self.assertFalse(os.path.exists(os.path.join(self.tempdir, 'file2')))

        application.execute()

        # Check that it has created two files
        self.assertTrue(os.path.exists(os.path.join(self.tempdir, 'file1')))
        self.assertTrue(os.path.exists(os.path.join(self.tempdir, 'file2')))

        # The above is reasonably sufficient to know that it works, given
        # that lower-level functionality has been tested in other unit tests.



#!/usr/bin/python
"""
cplot - contour plotting

Copyright (C) 2005-2011 National Technical University of Athens
Copyright (C) 2011 Antonis Christofides, Stefanos Kozanis

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

import os
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import cm

def plot_contours(filename, points, options):
    from scipy.interpolate import Rbf
    import numpy as np
    a = points
    GRANULARITY=options['granularity']
    (x0, y0, x1, y1) = (options['chart_bounds_bl_x'], 
                        options['chart_bounds_bl_y'],
                        options['chart_bounds_tr_x'], 
                        options['chart_bounds_tr_y'])
    dx, dy = (x1-x0)/GRANULARITY, (y1-y0)/GRANULARITY
    width, height = x1-x0, y1-y0

    # If we just interpolate, then the contours can be bad. What we actually
    # need to do is to tell the interpolator that far away the rainfall becomes
    # zero. So we actually work on a grid that is larger (three times larger),
    # we define rainfall at the edges as zero, we interpolate in there and
    # calculate the contours, and we crop it so that only its middle part is
    # shown. We call this grid the "greater" grid, so we suffix variables with
    # "gr" when they refer to it.
    # New update (2011-05-30 by Stefanos). The distance of boundary
    # can be altered by a factor (default value is 1) as well as
    # boundary value (default should be 0). Boundary mode 0 is for a
    # constant value, 1 for a ratio of the mean value
    bfac = options.get('boundary_distance_factor', 1)
    bvalue = options.get('boundary_value', 0)
    if options.get('boundary_mode', 0)==1:
        mean, count = 0, 0
        for p in a:
            if p:
                mean+=p[2]
                count+=1
        if count>0:
            mean/= count
            bvalue*= mean
    bwidth, bheight = bfac*width, bfac*height
    xgr0, ygr0, xgr1, ygr1 = (x0-bwidth, y0-bheight, x1+bwidth, y1+bheight)
    a.append((xgr0, ygr0, bvalue, 'virtual'))
    a.append((xgr1, ygr0, bvalue, 'virtual'))
    a.append((xgr0, ygr1, bvalue, 'virtual'))
    a.append((xgr1, ygr1, bvalue, 'virtual'))
    a.append((xgr0+bwidth/2, ygr0, bvalue, 'virtual'))
    a.append((xgr0+bwidth/2, ygr1, bvalue, 'virtual'))
    a.append((xgr0, ygr0+bheight/2, bvalue, 'virtual'))
    a.append((xgr1, ygr0+bheight/2, bvalue, 'virtual'))

    # Now interpolate
    x = [i[0] for i in a]
    y = [i[1] for i in a]
    z = [i[2] for i in a]
    rbfi = Rbf(x, y, z, function='linear')
    xx = np.arange(xgr0, xgr1+dx, dx)
    yy = np.arange(ygr0, ygr1+dy, dy)
    zz = np.empty([3*GRANULARITY+1, 3*GRANULARITY+1])
    for i in xrange(0, 3*GRANULARITY):
        for j in xrange(0, 3*GRANULARITY):
            zz[j][i] = rbfi(xx[i], yy[j])
    # Crop
    zz = zz[GRANULARITY:2*GRANULARITY+1, GRANULARITY:2*GRANULARITY+1]

    # Create the chart
    chart_large_dimension = float(options['chart_large_dimension'])
    chart_small_dimension = chart_large_dimension*min(y1-y0, x1-x0)/max(
                                                                y1-y0, x1-x0)
    if x1-x0>y1-y0:
        x_dim, y_dim = chart_large_dimension, chart_small_dimension
    else:
        x_dim, y_dim = chart_small_dimension, chart_large_dimension
    x_dim, y_dim = round(x_dim), round(y_dim)
    fig = plt.figure(figsize=(x_dim/100.0, y_dim/100.0), dpi=100.0)
    ax = plt.axes([0.0, 0.0, 1.0, 1.0]) # Axes should occupy no space
    ax.set_xticks([])
    ax.set_yticks([])
    plt.axis('off')
    for x, y, z, name in a:
        if name=='virtual': continue
        if options['draw_markers']:
            plt.plot(x, y, marker=options['markers_style'], linestyle='None',
                     color=options['markers_color'])
        if options['draw_labels']:
            plt.text(x, y, name, color=options['text_color'],
                     fontsize=options['labels_font_size'])
    args = [zz,]
    kwargs = {'interpolation': 'bilinear', 'origin': 'lower',
        'extent': (x0, x1, y0, y1)}
    if options['color_map']!='': 
        kwargs['cmap']=getattr(cm, options['color_map']) 
    im = plt.imshow(*args, **kwargs)
    if options['draw_contours']:
        cs = plt.contour(zz, extent=(x0, x1, y0, y1),
                         colors=options['contours_color'])
    plt.clabel(cs, inline=1, fontsize=options['contours_font_size'], fmt=options['labels_format'])

    fig.savefig(filename)

    if options['compose_background']:
        from PIL import Image, ImageChops
        contours = Image.open(filename)
        contours = contours.convert("RGB")
        cmethod = options['compose_method']
        size = contours.size
        background = Image.open(os.path.join(options['backgrounds_path'], options['background_image'])).convert("RGB")
        background_size = background.size
        if size!=background_size:
            background = background.resize(size)
        if cmethod=='composite':
            mask = Image.open(os.path.join(options['backgrounds_path'], options['mask_image']))
            mask_size = mask.size
            if size!=mask_size:
                mask = mask.resize(size)
        args = [background, contours]
        if options['swap_bg_fg']:
            args.reverse()
        if cmethod in ('blend', 'add', 'subtract'):
            args.append(options['compose_alpha'])
        if cmethod in ('add', 'subtract'):
            args.append(options['compose_offset'])
        if cmethod=='composite':
            args.append(mask)
        contours = getattr(ImageChops, cmethod)(*args)
        contours.save(filename)



from argparse import ArgumentParser
from datetime import datetime
import logging
import sys
import traceback

import six
from six.moves import configparser
from six.moves.configparser import RawConfigParser, NoOptionError


class InvalidOptionError(configparser.Error):
    pass


class WrongValueError(configparser.Error):
    pass


class CliApp(object):
    name = 'Replace me'
    description = 'Replace me'
                               # Section     Option            Default
    base_config_file_options = {'General': {'logfile':         '',
                                            'loglevel':        'WARNING',
                                            },
                                }
    config_file_options = {}
    cmdline_arguments = {}

    def read_command_line(self):
        parser = ArgumentParser(description=self.description)
        parser.add_argument('config_file', help='Configuration file')
        parser.add_argument('--traceback', action='store_true',
                            help='Display traceback on error')
        for arg in self.cmdline_arguments:
            parser.add_argument(arg, **(self.cmdline_arguments[arg]))
        self.args = parser.parse_args()

    def setup_logger(self):
        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(
            getattr(logging, self.config['General']['loglevel']))
        if self.config['General']['logfile']:
            self.logger.addHandler(
                logging.FileHandler(self.config['General']['logfile']))
        else:
            self.logger.addHandler(logging.StreamHandler())

    def read_configuration(self):
        # Assemble config_file_options
        for section in self.base_config_file_options:
            self.config_file_options[section].update(
                self.base_config_file_options[section])

        # Read config
        cp = RawConfigParser()
        if six.PY2:
            cp.read_file = cp.readfp
        cp.read_file(open(self.args.config_file))

        # Convert config to dictionary (for Python 2.7 compatibility)
        self.config = {}
        for section in cp.sections():
            self.config[section] = dict(cp.items(section))

        # Set defaults
        for section in self.config:
            section_options_key = section \
                if section in self.config_file_options else 'other'
            section_options = self.config_file_options[section_options_key]
            if section_options == 'nocheck':
                continue
            for option in section_options:
                value = section_options[option]
                if value is not None:
                    self.config[section].setdefault(option, value)

        # Check
        self.check_configuration()

    def check_configuration(self):
        # Check compulsory options and invalid options
        for section in self.config:
            optionsname = (section if section in self.config_file_options
                           else 'other')
            if self.config_file_options[optionsname] == 'nocheck':
                continue
            section_options = self.config_file_options[optionsname]
            for option in section_options:
                if (section_options[option] is None) and (
                        option not in self.config[section]):
                    raise NoOptionError(option, section)
            for option in self.config[section]:
                if option not in section_options:
                    raise InvalidOptionError(
                        'Invalid option {} in section [{}]'
                        .format(option, section))

        self.check_configuration_log_levels()

    def check_configuration_log_levels(self):
        log_levels = ['ERROR', 'WARNING', 'INFO', 'DEBUG']
        if self.config['General']['loglevel'] not in log_levels:
            raise WrongValueError('loglevel must be one of {}'.format(
                ', '.join(log_levels)))

    def run(self, dry=False):
        self.args = None
        self.logger = None
        try:
            self.read_command_line()
            self.read_configuration()
            self.setup_logger()
            self.logger.info(
                'Starting {}, {}'.format(self.name,
                                         datetime.today().isoformat()))
            if not dry:
                self.execute()
        except Exception as e:
            msg = str(e)
            sys.stderr.write(msg + '\n')
            if self.logger:
                self.logger.error(msg)
                self.logger.debug(traceback.format_exc())
            if self.args and self.args.traceback:
                raise
            sys.exit(1)
        finally:
            if self.logger:
                self.logger.info(
                    'Finished {}, {}'.format(self.name,
                                             datetime.today().isoformat()))

    def execute(self):
        raise NotImplementedError("CliApp is an abstract class")



from copy import copy
import os
import sys

import six

from pthelma.cliapp import CliApp, WrongValueError
from pthelma.timeseries import TimeStep, Timeseries, IntervalType


class FordoniaApp(CliApp):
    name = 'fordonia'
    description = 'Aggregate time series'
                          # Section     Option            Default
    config_file_options = {'General': {'base_dir':        None,
                                       'target_step':     None,
                                       'nominal_offset':  '0,0',
                                       'actual_offset':   '0,0',
                                       'missing_allowed': '0.0',
                                       'missing_flag':    'MISSING',
                                       },
                           'other':   {'source_file':     None,
                                       'target_file':     None,
                                       'interval_type':   None,
                                       },
                           }

    def read_configuration(self):
        super(FordoniaApp, self).read_configuration()

        # Save some stuff to easy to access variables
        genconfig = self.config['General']
        self.base_dir = genconfig['base_dir']
        self.missing_allowed = float(genconfig['missing_allowed'])
        minutes, months = [int(x.strip())
                           for x in genconfig['target_step'].split(',')]
        nominal_offset = [int(x.strip())
                          for x in genconfig['nominal_offset'].split(',')]
        actual_offset = [int(x.strip())
                         for x in genconfig['actual_offset'].split(',')]
        self.target_step = TimeStep(length_minutes=minutes,
                                    length_months=months,
                                    nominal_offset=nominal_offset,
                                    actual_offset=actual_offset)
        self.missing_flag = genconfig['missing_flag']

        # Convert time series sections into a list
        self.aggregation_items = []
        for section in self.config:
            if section in self.config_file_options:
                continue
            item = copy(self.config[section])
            item['interval_type'] = IntervalType.__dict__[
                item['interval_type'].upper()]
            self.aggregation_items.append(item)

    def check_configuration(self):
        super(FordoniaApp, self).check_configuration()

        self.check_configuration_missing_allowed()
        self.check_configuration_target_step()
        self.check_configuration_offset('nominal_offset')
        self.check_configuration_offset('actual_offset')
        for item in self.config:
            if item == 'General':
                continue
            self.check_configuration_item(item)

    def check_configuration_item(self, item):
        interval_type = self.config[item]['interval_type'].upper()
        if interval_type not in IntervalType.__dict__:
            raise WrongValueError('{} is not a valid interval type'.
                                  format(interval_type.lower()))

    def check_configuration_missing_allowed(self):
        missing_allowed = self.config['General']['missing_allowed']
        try:
            ma = float(missing_allowed)
        except ValueError:
            raise WrongValueError('{} is not a number'.format(missing_allowed))
        if ma < 0.0 or ma >= 1.0:
            raise WrongValueError('missing_allowed must be '
                                  'between 0 and 1')

    def check_configuration_target_step(self):
        target_step = self.config['General']['target_step']
        try:
            minutes, months = [int(x.strip()) for x in target_step.split(',')]
            if minutes and months:
                raise ValueError("not both minutes and months can be nonzero")
            if not minutes and not months:
                raise ValueError("one of minutes and months must be nonzero")
        except (IndexError, ValueError):
            raise WrongValueError(
                '"{}" is not an appropriate time step; use "X,Y" where X is '
                'minutes and Y is months; one and only one must be nonzero.'
                .format(target_step))

    def check_configuration_offset(self, offsetname):
        offset = self.config['General'][offsetname]
        try:
            minutes, months = [int(x.strip()) for x in offset.split(',')]
        except (IndexError, ValueError):
            raise WrongValueError(
                '"{}" is not an appropriate offset; use "X,Y" where X is '
                'minutes and Y is months.'.format(offset))

    def execute(self):
        if self.base_dir:
            os.chdir(self.base_dir)
        for item in self.aggregation_items:
            self.logger.info('Processing {}'.format(item['source_file']))
            try:
                self.execute_item(item)
            except Exception as e:
                exc_class, exc, tb = sys.exc_info()
                new_exception = Exception(
                    'A {} occurred while processing {}: {}'.format(
                        exc_class.__name__,
                        os.path.join(self.base_dir, item['source_file']),
                        str(e)))
                if six.PY2:
                    exec('raise new_exception.__class__, new_exception, tb')
                else:
                    raise new_exception.with_traceback(tb)

    def execute_item(self, item):
        source_ts = Timeseries()
        with open(item['source_file']) as f:
            source_ts.read_file(f)
        self.target_step.interval_type = item['interval_type']
        target_ts, missing = source_ts.aggregate(
            self.target_step,
            missing_allowed=self.missing_allowed,
            missing_flag=self.missing_flag)
        with open(item['target_file'], 'w') as f:
            target_ts.write_file(f, version=3)



import codecs
from copy import copy
from datetime import datetime, timedelta
import os

from six import StringIO

import requests
from requests.exceptions import HTTPError

from pthelma import enhydris_api
from pthelma.cliapp import CliApp, WrongValueError
from pthelma.timeseries import Timeseries


class TimeseriesCache(object):

    def __init__(self, timeseries_group):
        self.timeseries_group = timeseries_group

    def update(self):
        for item in self.timeseries_group:
            self.base_url = item['base_url']
            if self.base_url[-1] != '/':
                self.base_url += '/'
            self.timeseries_id = item['id']
            self.user = item['user']
            self.password = item['password']
            self.filename = item['file']
            self.update_for_one_timeseries()

    def update_for_one_timeseries(self):
        ts1 = self.read_timeseries_from_cache_file()
        end_date = self.get_timeseries_end_date(ts1)
        start_date = end_date + timedelta(minutes=1)
        self.append_newer_timeseries(start_date, ts1)
        with codecs.open(self.filename, 'w', 'utf-8') as f:
            ts1.write_file(f, version=3)

    def read_timeseries_from_cache_file(self):
        result = Timeseries()
        if os.path.exists(self.filename):
            with open(self.filename) as f:
                try:
                    result.read_file(f)
                except ValueError:
                    # File may be corrupted; continue with empty time series
                    result = Timeseries()
        return result

    def get_timeseries_end_date(self, timeseries):
        try:
            end_date = timeseries.bounding_dates()[1]
        except TypeError:
            # Timeseries is totally empty; no start and end date
            end_date = datetime(1, 1, 1, 0, 0)
        return end_date

    def append_newer_timeseries(self, start_date, ts1):
        self.session_cookies = enhydris_api.login(self.base_url, self.user,
                                                  self.password)
        url = self.base_url + 'timeseries/d/{}/download/{}/?version=3'.format(
            self.timeseries_id, start_date.isoformat())
        r = requests.get(url, cookies=self.session_cookies)
        if r.status_code != 200:
            raise HTTPError('Error {} while getting {}'.format(r.status_code,
                                                               url))
        responseio = StringIO(r.text)
        ts2 = Timeseries()
        ts2.read_file(responseio)
        responseio.seek(0)
        ts1.read_meta(responseio)
        ts1.append(ts2)


class PondApp(CliApp):
    name = 'pond'
    description = 'Local filesystem cache of Enhydris data'
                          # Section     Option      Default
    config_file_options = {'General': {'cache_dir': os.getcwd()
                                       },
                           'other':   {'base_url':  None,
                                       'id':        None,
                                       'user':      '',
                                       'password':  '',
                                       'file':      None,
                                       },
                           }

    def check_configuration(self):
        super(PondApp, self).check_configuration()

        for section in self.config:
            if section == 'General':
                continue
            try:
                int(self.config[section]['id'])
            except ValueError as e:
                raise WrongValueError('"{}" is not a valid integer'.format(
                    self.config[section]['id']))

    def read_configuration(self):
        super(PondApp, self).read_configuration()

        # Convert all sections but 'General' into a list of time series
        self.timeseries_group = []
        for section in self.config:
            if section == 'General':
                continue
            item = copy(self.config[section])
            item['name'] = section
            item['id'] = int(item['id'])
            self.timeseries_group.append(item)

    def execute(self):
        os.chdir(self.config['General']['cache_dir'])
        self.cache = TimeseriesCache(self.timeseries_group)
        self.cache.update()



#!/usr/bin/env python

import atexit
from compileall import compile_file
from ctypes import CDLL, c_char_p
import os
import platform
import shutil
import sys
import tempfile
import textwrap
try:
    # Python 3
    from urllib.request import urlopen
except ImportError:
    # Python 2
    from urllib2 import urlopen
from zipfile import ZipFile

import requests
from setuptools import setup, find_packages
from pkg_resources import parse_version


REQUIRED_DICKINSON_VERSION = '0.2.0'


def add_dir_to_zipfile(zipfile, dir, zipdir):
    """Recursively adds all files of dir in the zipfile, modifying their
    location by replacing dir with zipdir."""
    for dirpath, dirnames, filenames in os.walk(dir):
        assert(dirpath.startswith(dir))
        zippath = dirpath.replace(dir, zipdir, 1)
        for f in filenames:
            zipfile.write(os.path.join(dirpath, f), os.path.join(zippath, f))


class VersionError(Exception):
    pass


try:
    dickinson = CDLL('dickinson.dll'
                     if platform.system() == 'Windows'
                     else 'libdickinson.so')
    dickinson_version = c_char_p.in_dll(dickinson, 'dickinson_version'
                                        ).value.decode('ascii')
    if dickinson_version != 'dev' and parse_version(dickinson_version) \
            < parse_version(REQUIRED_DICKINSON_VERSION):
        raise VersionError('Too old version of dickinson: {0}'.format(
                           dickinson_version))
except OSError as e:
    sys.stderr.write(str(e) + '\n')
    sys.stderr.write(textwrap.dedent('''
        Please make sure you have installed dickinson
        (see http://dickinson.readthedocs.org/).
        '''))
    sys.exit(1)
except ValueError as e:
    sys.stderr.write(str(e) + '\n')
    sys.stderr.write(textwrap.dedent('''
        Apparently the version of dickinson that is installed doesn't export
        a dickinson_version variable. Maybe it's too old. Please install at
        least version {0} (see http://dickinson.readthedocs.org/).
        '''.format(REQUIRED_DICKINSON_VERSION)))
    sys.exit(1)
except VersionError as e:
    sys.stderr.write(str(e) + '\n')
    sys.stderr.write(textwrap.dedent('''
        Please install at least version {0} of dickinson
        (see http://dickinson.readthedocs.org/)
        '''.format(REQUIRED_DICKINSON_VERSION)))
    sys.exit(1)


installation_requirements = ["pytz",
                             "simpletail>=0.1.1",
                             "requests>=1,<3",
                             "six>=1.6,<2",
                             ]
if sys.platform != 'win32':
    installation_requirements.extend(["numpy>=1.5,<2",
                                      "gdal>=1.9,<2",
                                      ])
else:
    installation_requirements.extend(["pyodbc>=3,<4",
                                      ])

kwargs = {
    'name': "pthelma",
    'version': "0.9.0",
    'license': "GPL3",
    'description': "Hydro/meteorological timeseries library",
    'author': "Antonis Christofides",
    'author_email': "anthony@itia.ntua.gr",
    'packages': find_packages(),
    'scripts': ['bin/loggertodb', 'bin/bitia', 'bin/pond', 'bin/fordonia'],
    'test_suite': "tests",
    'install_requires': installation_requirements,
    'options': {'py2exe': {'includes': ['pyodbc',

                                        # The following two are required by
                                        # pyodbc but might not be found by
                                        # py2exe, which doesn't look in C++
                                        # files.
                                        'decimal',
                                        'datetime',

                                        # ConfigParser is not automatically
                                        # found by py2exe when imported as
                                        # six.moves.configparser.
                                        'ConfigParser',
                                        ],
                           'excludes': ['spatial']}},
}

if len(sys.argv) >= 2 and sys.argv[1] == 'py2exe':
    import py2exe
    py2exe  # Does nothing, but lint checkers won't warn about unused py2exe

    # Download and save MSVC++ 2008 redistributable in a temporary directory
    # that will be removed when the program exits
    response = urlopen('http://download.microsoft.com/download/'
                       '1/1/1/1116b75a-9ec3-481a-a3c8-1777b5381140/'
                       'vcredist_x86.exe')
    tmpdir = tempfile.mkdtemp()
    atexit.register(lambda: shutil.rmtree(tmpdir, ignore_errors=True))
    vcredist_filename = os.path.join(tmpdir, 'vcredist_x86.exe')
    with open(vcredist_filename, 'wb') as f:
        f.write(response.read())

    # Add the MSVC++ 2008 redistributable and dickinson to the produced files
    kwargs['data_files'] = [('', [r'\Windows\System32\dickinson.dll',
                                  vcredist_filename])]

    # Specify program executable
    kwargs['console'] = ['bin/loggertodb', 'bin/pond', 'bin/fordonia']

    # python-requests' cacert.pem.
    # First, we need to add cacert.pem to the bundle. That's the easy part.
    # However, requests.certs looks for cacert.pem in the directory from which
    # requests.certs has been loaded (i.e.
    # os.path.dirname(requests.certs.__file__). This is probably wrong; it
    # should be using pkg_resources to load it. The problem is that py2exe puts
    # requests in library.zip, so attempting to read stuff from the above
    # directory won't work.  So, instead, we add cacert.pem to the bundle,
    # deploy it alongside library.zip, and modify requests/certs.py at bundle
    # creation time so that, when deployed, it will read certs.py from the
    # bundle location.
    cacert_filename = os.path.join(os.path.dirname(requests.__file__),
                                   'cacert.pem')
    kwargs['data_files'][0][1].append(cacert_filename)
    certs_path = os.path.join(
        os.path.dirname(requests.__file__),
        'certs.py')  # We can't use requests.certs.__file__; it may be the .pyc
    try:
        os.remove(certs_path + '.bak')
    except:
        pass
    os.rename(certs_path, certs_path + '.bak')

    def restore_certs():
        os.remove(certs_path)
        os.rename(certs_path + '.bak', certs_path)
        compile_file(certs_path, force=True)

    atexit.register(restore_certs)
    with open(certs_path, 'w') as f:
        f.write(textwrap.dedent("""\
            import os
            import sys

            def where():
                # This is not the function packaged by python-requests, but
                # a patch made by py2exe setup. If you are looking at the code
                # bundled by py2exe (unlikely, since it normally doesn't
                # include the source), this is normal. If, however, you are
                # looking at the python-requests code installed on the system
                # where you run py2exe, it is likely that something went wrong
                # and py2exe failed to revert this file to the original, as it
                # should have done after creating the bundle. You should delete
                # and re-install python-requests. See pthelma's setup.py for
                # more information.
                return os.path.join(os.path.dirname(sys.executable),
                                    'cacert.pem')
            """))
    compile_file(certs_path, force=True)

setup(**kwargs)


if len(sys.argv) >= 2 and sys.argv[1] == 'py2exe':
    # Add pytz zoneinfo to library.zip
    import pytz
    zoneinfo_dir = os.path.join(os.path.dirname(pytz.__file__), 'zoneinfo')
    with ZipFile(os.path.join('dist', 'library.zip'), 'a') as z:
        add_dir_to_zipfile(z, zoneinfo_dir, os.path.join('pytz', 'zoneinfo'))



from datetime import datetime
from six import StringIO

import requests

from pthelma.timeseries import datetime_from_iso


def urljoin(*args):
    result = '/'.join([s.strip('/') for s in args])
    if args[-1].endswith('/'):
        result += '/'
    return result


# My understanding from requests' documentation is that when I make a post
# request, it shouldn't be necessary to specify Content-Type:
# application/x-www-form-urlencoded, and that requests adds the header
# automatically. However, when running in Python 3, apparently requests does
# not add the header (although it does convert the post data to
# x-www-form-urlencoded format). This is why the header has been explicitly
# specified in all post requests.

def login(base_url, username, password):
    if not username:
        return {}
    login_url = urljoin(base_url, 'accounts/login/')
    r = requests.get(login_url)
    r.raise_for_status()
    r1 = requests.post(
        login_url,
        headers={'X-CSRFToken': r.cookies['csrftoken'],
                 'Referer': login_url,
                 'Content-Type': 'application/x-www-form-urlencoded'
                 # See comment about x-www-form-urlencoded above
                 },
        data='username={}&password={}'.format(username, password),
        cookies=r.cookies,
        allow_redirects=False)
    r1.raise_for_status()
    result = r1.cookies
    return result


def get_model(base_url, session_cookies, model, id):
    url = urljoin(base_url, 'api/{}/{}/'.format(model, id))
    r = requests.get(url, cookies=session_cookies)
    r.raise_for_status()
    return r.json()


def post_model(base_url, session_cookies, model, data):
    r = requests.post(
        urljoin(base_url, 'api/{}/'.format(model)),
        headers={'X-CSRFToken': session_cookies['csrftoken'],
                 'Content-Type': 'application/x-www-form-urlencoded'
                 # See comment about x-www-form-urlencoded above
                 },
        cookies=session_cookies,
        data=data
    )
    r.raise_for_status()
    return r.json()['id']


def delete_model(base_url, session_cookies, model, id):
    url = urljoin(base_url, 'api/{}/{}/'.format(model, id))
    r = requests.delete(url, cookies=session_cookies,
                        headers={'X-CSRFToken': session_cookies['csrftoken']})
    if r.status_code != 204:
        raise requests.exceptions.HTTPError()


def read_tsdata(base_url, session_cookies, ts):
    r = requests.get(base_url + 'api/tsdata/{0}/'.format(ts.id),
                     cookies=session_cookies)
    r.raise_for_status()
    ts.read(StringIO(r.text))


def post_tsdata(base_url, session_cookies, timeseries):
    f = StringIO()
    timeseries.write(f)
    url = urljoin(base_url, 'api/tsdata/{}/'.format(timeseries.id))
    r = requests.post(
        url,
        data={'timeseries_records': f.getvalue()},
        headers={'Content-type': 'application/x-www-form-urlencoded',
                 'X-CSRFToken': session_cookies['csrftoken']},
        cookies=session_cookies)
    r.raise_for_status()
    return r.text


def get_ts_end_date(base_url, session_cookies, ts_id):
    r = requests.get('{}timeseries/d/{}/bottom/'.format(base_url, ts_id),
                     cookies=session_cookies)
    r.raise_for_status()
    lines = r.text.splitlines()
    lines.reverse()
    for line in [x.strip() for x in lines]:
        if not line:
            continue
        datestring = line.split(',')[0]
        return datetime_from_iso(datestring)
    return datetime(1, 1, 1)



#!/usr/bin/python
"""
meteocalcs - meteorological calculations such as Heat Index, Wind
Chill, Evaporation etc.

Copyright (C) 2005-2011 National Technical University of Athens
Copyright (C) 2011 Stefanos Kozanis

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

from math import exp, isnan

#Gravitational acceleration at phi=45deg m^2/s
g0 = 9.80665
#Universal gas constant for air in N.m/(mol.K)
Rs = 8.31432
#Molar mass of Earth's air in kg/mol
Mair = 0.0289644


def HeatIndex(Tc, RH):
    """Return the Heat Index in degrees C. Tc in degrees C,
    RH rel. hum. in % (e.g. 40%). When Tc<26.7
    returns the actual Tc.
    """
    if isnan(Tc) or isnan(RH):
        return float('NaN')
    if Tc<26.7:
        return Tc
    Tf = float(Tc)*9/5 + 32
    HI = -42.379 + 2.04901523*Tf + 10.14333127*RH - 0.22475541*Tf*RH -\
         6.83783e-3* Tf**2 - 5.481717e-2* RH**2 +\
         1.22874e-3* Tf**2 * RH + 8.5282e-4* Tf* RH**2 -\
         1.99e-6* Tf**2 * RH**2
    return (HI-32)*5/9

def SSI(Tc, RH):
    """Return the Summer Simmer Index in degrees C. Tc in degrees C,
    RH rel. hum. in % (e.g. 40%). When Tc<22
    returns the actual Tc.
    """
    if isnan(Tc) or isnan(RH):
        return float('NaN')
    if Tc<22:
        return Tc
    Tf = float(Tc)*9/5 + 32
    ssi =  1.98*(Tf-(0.55-0.0055*RH)*(Tf-58))-56.83
    return (ssi-32)*5/9

def IDM(T, Precip, is_annual=False):
    """Return the de Martonne Aridity Index. T in degrees C,
    Precip in mm. If is_annual=False values should be monthly
    or else annual.
    """
    if isnan(T) or isnan(Precip):
        return float('NaN')
    f=1 if is_annual else 12
    return f*Precip/(T+10)

def BarometricFormula(T, Pb, hdiff):
    """Return the barometric pressure at a level
    h if the pressure Pb at an altutde hb is given for
    atmospheric temperature T, according to the
    barometric formula. hdiff is h-hb.
    """
    for v in (T, Pb, hdiff):
        if isnan(v):
            return float('NaN')
            break
    T+= 273.75
    return Pb * exp( (-g0*Mair*hdiff)/(Rs * T) )
